import argparse
import os
import pandas
import re
import shutil
import sys
import tempfile
import zipfile
from collections import OrderedDict
from enum import Enum

from schrodinger import structure
from schrodinger.job import jobcontrol as jc
from schrodinger.job import launcher
from schrodinger.job.util import hunt
from schrodinger.utils import subprocess
from schrodinger.utils import fileutils

import KNIME_workflow_wrapper as kwr


class PrintType(Enum):
    NO_PRINT = 0
    USAGE = 1
    COMMAND = 2
    STU_COMMAND = 3


# Global variables
USAGE = '''
%(prog)s workflow (-print_usage|-print_cmd|-print_knime_cmd|-stu|-run|-GUI)
[KNIME options] [Script specific input, output and parameter arguments]
or
%(prog)s -printallowed
'''
SCRIPT_DESCRIPTION = '''
Interface script to run batch executions in an user friendly manner'''

parser = None
in_maestro = False
tempdir = tempfile.gettempdir()

# List of option keywords
INPUT_MAE = '-istruct'
INPUT_ALI = '-ialign'
INPUT_SEQ = '-iseq'
INPUT_CSV = '-itxt'
INPUT_SDF = '-isdf'
INPUT_MOL2 = '-imol2'
INPUT_SMI = '-ismi'
INPUT_PDB = '-ipdb'
INPUT_XLS = '-ixls'
INPUT_FP = '-ifp'
INPUT_GRD = '-igrd'
INPUT_HYP = '-ihyp'
INPUT_QSAR = '-iqzip'
INPUT_CMS = '-icms'
INPUT_FEP = '-ifmp'
INPUT_LRSDF = '-ilr'
INPUT_KNIME_MODEL = '-ikmodel'
INPUT_KNIME_LISTFILES = '-ilistfiles'
INPUT_KNIME_TABLE_READER = '-itable'

OUTPUT_MAE = '-ostruct'
OUTPUT_ALI = '-oalign'
OUTPUT_SEQ = '-oseq'
OUTPUT_CSV = '-otxt'
OUTPUT_SDF = '-osdf'
OUTPUT_MOL2 = '-omol2'
OUTPUT_XLS = '-oxls'
OUTPUT_FP = '-ofp'
OUTPUT_GRD = '-ogrd'
OUTPUT_HYP = '-ohyp'
OUTPUT_QSAR = '-oqzip'
OUTPUT_KNIME_TABLE_WRITER = '-otable'

NODE_DATA = {
    kwr.MOLECULE_READER: {
        'option': INPUT_MAE,
        'count': 0
    },
    kwr.ALIGNMENT_READER: {
        'option': INPUT_ALI,
        'count': 0
    },
    kwr.SEQUENCE_READER: {
        'option': INPUT_SEQ,
        'count': 0
    },
    kwr.CSV_READER: {
        'option': INPUT_CSV,
        'count': 0
    },
    kwr.SDF_READER: {
        'option': INPUT_SDF,
        'count': 0
    },
    kwr.MOL2_READER: {
        'option': INPUT_MOL2,
        'count': 0
    },
    kwr.SMILES_READER: {
        'option': INPUT_SMI,
        'count': 0
    },
    kwr.PDB_READER: {
        'option': INPUT_PDB,
        'count': 0
    },
    kwr.EXCEL_READER: {
        'option': INPUT_XLS,
        'count': 0
    },
    kwr.FP_READER: {
        'option': INPUT_FP,
        'count': 0
    },
    kwr.GRID_READER: {
        'option': INPUT_GRD,
        'count': 0
    },
    kwr.HYP_READER: {
        'option': INPUT_HYP,
        'count': 0
    },
    kwr.QSAR_READER: {
        'option': INPUT_QSAR,
        'count': 0
    },
    kwr.CMS_READER: {
        'option': INPUT_CMS,
        'count': 0
    },
    kwr.FEP_READER: {
        'option': INPUT_FEP,
        'count': 0
    },
    kwr.LD_INPUT: {
        'option': INPUT_LRSDF,
        'count': 0
    },
    kwr.KNIME_MODEL_READER: {
        'option': INPUT_KNIME_MODEL,
        'count': 0
    },
    kwr.KNIME_LIST_FILES: {
        'option': INPUT_KNIME_LISTFILES,
        'count': 0
    },
    kwr.KNIME_TABLE_READER: {
        'option': INPUT_KNIME_TABLE_READER,
        'count': 0
    },
    kwr.MOLECULE_WRITER: {
        'option': OUTPUT_MAE,
        'count': 0
    },
    kwr.ALIGNMENT_WRITER: {
        'option': OUTPUT_ALI,
        'count': 0
    },
    kwr.SEQUENCE_WRITER: {
        'option': OUTPUT_SEQ,
        'count': 0
    },
    kwr.CSV_WRITER: {
        'option': OUTPUT_CSV,
        'count': 0
    },
    kwr.SDF_WRITER: {
        'option': OUTPUT_SDF,
        'count': 0
    },
    kwr.MOL2_WRITER: {
        'option': OUTPUT_MOL2,
        'count': 0
    },
    kwr.EXCEL_WRITER: {
        'option': OUTPUT_XLS,
        'count': 0
    },
    kwr.FP_WRITER: {
        'option': OUTPUT_FP,
        'count': 0
    },
    kwr.GRID_WRITER: {
        'option': OUTPUT_GRD,
        'count': 0
    },
    kwr.HYP_WRITER: {
        'option': OUTPUT_HYP,
        'count': 0
    },
    kwr.QSAR_WRITER: {
        'option': OUTPUT_QSAR,
        'count': 0
    },
    kwr.KNIME_TABLE_WRITER: {
        'option': OUTPUT_KNIME_TABLE_WRITER,
        'count': 0
    },
    kwr.CSV_READER_1: {
        'option': INPUT_CSV,
        'count': 0
    },
    kwr.EXCEL_READER_1: {
        'option': INPUT_XLS,
        'count': 0
    },
    kwr.KNIME_TABLE_READER_1: {
        'option': INPUT_KNIME_TABLE_READER,
        'count': 0
    },
    kwr.KNIME_MODEL_READER_1: {
        'option': INPUT_KNIME_MODEL,
        'count': 0
    },
    kwr.KNIME_LIST_FILES_1: {
        'option': INPUT_KNIME_LISTFILES,
        'count': 0
    },
    kwr.CSV_WRITER_1: {
        'option': OUTPUT_CSV,
        'count': 0
    },
    kwr.EXCEL_WRITER_1: {
        'option': OUTPUT_XLS,
        'count': 0
    },
    kwr.KNIME_TABLE_WRITER_1: {
        'option': OUTPUT_KNIME_TABLE_WRITER,
        'count': 0
    }
}

grid_hyp_usage = '''
    Note:
    The input hypothesis to %(hyp_opt)s have to be specified without any
    extensions, like:
    /path/to/hypothesis/hypo_prefix
''' % {
    'hyp_opt': INPUT_HYP
}

VALUE = 'value'
TEXT = 'text'
FROM_LR = 'from_lr'
QUOTE = 'quote'

KNIME_WORKFLOW_FILE = "workflow.knime"

SCRIPT_PATH = "$SCHRODINGER/run KNIME_batch.py"
STU_ARGS = " -reset -suppressErrors -stdout INFO -stderr DEBUG " +\
            "-data workspace -maxThreads 1 -schrodingerTempDir " +\
            "${CWD}/schrodingerTempDir -knimeTempDir ${CWD}/knimeTempDir " +\
            "-deleteTempFiles false"


def reset_node_counts():
    for node in NODE_DATA.values():
        node['count'] = 0


def has_option_in_nodedata(option):
    for node in NODE_DATA.values():
        if option.startswith(node['option']):
            return True

    return False


def is_infile_option(option):
    return option.startswith('-i') and has_option_in_nodedata(option)


def is_outfile_option(option):
    return option.startswith('-o') and has_option_in_nodedata(option)


def validate_workflow_path(workflow):
    if not os.path.exists(workflow):
        raise ValueError("Specified workflow '%s' does not exist." % (workflow))

    if os.path.isfile(workflow) and not zipfile.is_zipfile(workflow):
        raise ValueError("Specified workflow file '%s' is not a zip file." %
                         (workflow))

    return os.path.abspath(workflow)


def unzip_workflow(workflow_file, tempdir):
    '''
    Unzip the KNIME workflow file under temporary directory and return the
    unzipped directory path. If workflow is not a local zip file (directory
    or a remote workflow path), just return the workflow_file

    :param workflow_file: Path to the workflow file (.knwf)
    :type workflow_file: `str`

    :return: Input workflow file (unaltered) or unzipped directory
    :rtype: `str`
    '''
    if not zipfile.is_zipfile(workflow_file):
        return workflow_file

    with zipfile.ZipFile(workflow_file) as zf:
        # Extract to the cwd
        workflow_dir = zf.namelist()[0].split('/')[0]
        zf.extractall(tempdir)

    return os.path.join(tempdir, workflow_dir)


#*******************************************************************************
class KNIMEBatch:

    def __init__(self, workflow, debug):

        self.workflow = validate_workflow_path(workflow)
        self.logger = kwr.create_logger(self.workflow, in_maestro, debug)

        # Create a temporary directory
        self.tempdir = tempfile.mkdtemp(prefix="KNIME_")
        self.logger.info("Temporary directory is: " + self.tempdir)

        # If workflow is a zipfile, extract it to self.tempdir
        # KnimeWorkflowParser class only handles workflow dirs and not knwf
        # files
        processed_workflow_path = unzip_workflow(self.workflow, self.tempdir)
        self.logger.debug("Processed workflow location:" +
                          processed_workflow_path)

        # Instance of the class
        self.workflow_parser = kwr.KnimeWorkflowParser(processed_workflow_path,
                                                       self.tempdir,
                                                       self.logger)
        # Allowed input, output and GUI parameter nodes
        self.workflow_parser.lookUpTables()
        # Parse workflow and create the node and parameter lists
        self.workflow_parser.parseWorkflow()

        # option_map is a map of option_name to a list with 2 elements
        # [node_id, value]
        self.option_map = {}

    def _quote_string(self, arg):
        '''
        Quote string if it has spaces and not already quoted
        '''
        if (' ' in arg) and not \
            (re.search('^".*"$', arg) or re.search("^'.*'$", arg)):
            arg = "\"" + arg + "\""
        return arg

    def getInOutOptions(self, node_list, print_type=PrintType.NO_PRINT):
        '''
        For the given list of input or output nodes, it returns a tuple:
        (message, command, option_map)
        message - Usage message to be printed
        command - command options to be printed
        option_map - map of option to a map of values for that option::

            {option: {'value': <opt_value>, 'text': <node's label text>,
                      'from_lr': True/False}}
        '''
        message = ""
        sample_cmd = ""
        option_map = OrderedDict()
        reset_node_counts()

        for node in node_list:
            node_type = node['nodeType'].strip()
            option = NODE_DATA[node_type]['option']
            if NODE_DATA[node_type]['count'] == 0:
                opt_name = option
            else:
                opt_name = option + str(NODE_DATA[node_type]['count'])
            NODE_DATA[node_type]['count'] += 1

            opt_val = node['fileName']
            node_name = node['nodeName']
            node_type = node['nodeType']

            # from_livereport is True if this options value is taken from
            # livereport. Say for a SDF Reader or CSV Reader with tag
            # 'livereport' or a 'LiveDesign Input' node
            from_livereport = node_type == kwr.LD_INPUT or (
                node_type in kwr.LIVEREPORT_INPUT_NODES) and bool(
                    kwr.RE_LIVEREPORT.search(node_name))

            message += "\t%s <file> [Default: %s]\n" % (
                opt_name, self._quote_string(opt_val))

            if node_name:
                message += "\t%s\n" % node_name

            if print_type == PrintType.STU_COMMAND:
                stu_file = f"${{CWD}}/{os.path.basename(os.path.abspath(opt_val))}"
                sample_cmd += f" {opt_name} {self._quote_string(stu_file)}"
            else:
                sample_cmd += f" {opt_name} {self._quote_string(opt_val)}"

            node_id = ""
            if node['location'] != '':
                node_id = node['location'] + "/"
            node_id += node['nodeNumber']
            # The second item in the list will be filled in execute()
            self.option_map[opt_name] = [node_id, None]

            # Populate option_map
            values_map = {}
            values_map[VALUE] = opt_val
            values_map[TEXT] = node_name
            values_map[FROM_LR] = from_livereport
            option_map[opt_name] = values_map

        return (message, sample_cmd, option_map)

    #***************************************************************************
    def getParameterOptions(self, node_list):
        '''
        For the given list of GUI parameter nodes, it returns a tuple:
        (message, command, option_map)
        message - Usage message to be printed
        command - command options to be printed
        option_map - map of option to a map of values for that option, like
        {option: {'value': <opt_value>, 'text': <node's label text>}}
        '''
        message = ""
        sample_cmd = ""
        option_map = OrderedDict()

        for node in node_list:
            possible_values = ""
            param_dict = node['parameters'][0]
            opt_name = "-" + param_dict['varName'].replace(" ", "_")
            default_value = param_dict['value']

            message += "\t%s <Name:%s, Label:%s> [Default: %s]" % \
                       (opt_name,
                        node['nodeName'],
                        node['nodeLabel'].rstrip(":"),
                        self._quote_string(default_value))

            sample_cmd += f" {opt_name} {self._quote_string(default_value)}"

            if param_dict['possible values'] != []:
                possible_values = ",".join(param_dict['possible values'])
                message += " {%s}\n" % possible_values
            else:
                message += "\n"

            message += "\t %s\n" % node['nodeDescription']

            node_id = ""
            if node['location'] != '':
                node_id = node['location'] + "/"
            node_id += node['nodeNumber']
            # The second item in the list will be filled in execute()
            self.option_map[opt_name] = [node_id, None]

            values_map = {}
            values_map[VALUE] = default_value
            values_map[TEXT] = node['nodeLabel']
            values_map[QUOTE] = node['input_quoted']
            option_map[opt_name] = values_map

        return (message, sample_cmd, option_map)

    #***************************************************************************
    def getHelpMessage(self, print_type: PrintType):
        message = self.workflow_parser.workflowName.replace('_', ' ') + "\n"
        message += self.workflow_parser.workflowDescription
        message += create_parser().format_help()

        if print_type == PrintType.STU_COMMAND:
            workflow = f'${{CWD}}/{os.path.basename(self.workflow)}'
            (base, ext) = os.path.splitext(workflow)
            sample_cmd = f'{SCRIPT_PATH} {self._quote_string(workflow)} -run {STU_ARGS} -destFile={base}-executed{ext}'
        else:
            sample_cmd = f'{SCRIPT_PATH} "{self.workflow}" -run'

        opt_name = ""

        # Inputs
        message += "\nInputs:\n"

        if self.workflow_parser.input_nodes == []:
            message += "\tNone\n"
        else:
            (message_part, sample_cmd_part,
             _) = self.getInOutOptions(self.workflow_parser.input_nodes,
                                       print_type)
            message += message_part
            sample_cmd += sample_cmd_part

            message += grid_hyp_usage + "\n"

        # Outputs
        message += "Outputs:\n"

        if self.workflow_parser.output_nodes == []:
            message += "\tNone\n"
        else:
            (message_part, sample_cmd_part,
             _) = self.getInOutOptions(self.workflow_parser.output_nodes,
                                       print_type)
            message += message_part
            sample_cmd += sample_cmd_part

        # Parameters
        message += "Parameters:\n"
        if self.workflow_parser.GUI_parameters == []:
            message += "\tNone\n"
        else:
            (message_part, sample_cmd_part,
             _) = self.getParameterOptions(self.workflow_parser.GUI_parameters)
            message += message_part
            sample_cmd += sample_cmd_part

        if print_type in (PrintType.STU_COMMAND, PrintType.COMMAND):
            return sample_cmd
        elif print_type == PrintType.USAGE:
            return f"{message}\nCommand example:\n{sample_cmd}"

    #***************************************************************************
    @staticmethod
    def _replaceOptionInCmd(opt_key, value, option_str):
        '''
        Replace the -option's value based on the value passed in the commandline
        Ex: -option=1,DataURL/0,file:/tmp/1648/in.mae,String

        :type opt_key: string
        :param opt_key: The command line option, like, -imae, -igrd etc...

        :type value: str
        :param value: The opt_key's value specified in the commandline

        :type option_str: str
        :param option_str: '-option=' string from the KNIME batch commandline

        :rtype: str
        :return: The replaced '-option=' based on the value passed
        '''

        option_list = option_str.split(",")
        option_list[2] = value
        # KNIME-3948 Handle commas in option value
        # The template for the -option is:
        # -option=<node_id>,<attribute>,<value>,<type>
        # Ex: "-option=17,value/value,view1,view2,view3,String"
        # Comma is the separator for the '-option' but as per above example
        # there could be commas in the <value> part which has to be handled
        # separately. Other than the <value> part, other parts wont have any
        # commas in itself. So, if the <value> part has commas in itself we
        # can throw away everything and simply replace the same with 'value'
        # argument.
        if len(option_list) > 4:
            # If more than 4 elements, put last element in index 3
            # and remove rest
            option_list[3] = option_list[-1]
            del option_list[4:]

        return ','.join(option_list)

    def execute(self, args_list, launch_job=False, exec_cmd=True):
        self.getHelpMessage(PrintType.NO_PRINT)
        options, other_args = self.redo_parseargs(args_list)

        host = None
        if os.environ['SCHRODINGER_NODELIST']:
            host, cpus = jc.get_command_line_host_list()[0]

        # Prepare the command
        run = os.path.join(os.environ['SCHRODINGER'], "run")
        # script to run is KNIME_workflow_wrapper.py
        cmd = [run, self.workflow_parser.script_name]

        # workflow path should be the first argument for
        # KNIME_workflow_wrapper.py.
        workflow = os.path.basename(self.workflow) if host else self.workflow
        cmd.append(workflow)

        cmd.append("-NOJOBID")

        input_files = []
        output_files = []
        # KNIME-5695
        # Based on -HOST, add -primaryHost and -primaryCPU options so that KNIME
        # nodes in the workflow will honor them
        if host:
            cmd.extend(['-primaryHost', host])
            cpus = cpus if cpus else 1
            cmd.extend(['-primaryCPU', str(cpus)])

        cmd = self.workflow_parser.addInputData(cmd, in_maestro)
        cmd = self.workflow_parser.addOutputData(cmd)
        cmd = self.workflow_parser.addGUIparameters(cmd)

        # Replace values passed in cmdline
        pos = 0
        for arg in cmd:
            # option_list is the -option arg split into a list on commas
            # Ex: -option=12,filename,/tmp/1648/out.mol2,String
            option_list = arg.split(",")
            if (option_list[0].startswith("-option=")):
                node_id = option_list[0].split("=")[1]
                if node_id != '':
                    for option_name in self.option_map:
                        map_node_id, map_node_value = self.option_map[
                            option_name]
                        if map_node_id == node_id and map_node_value is not None:
                            if host and launch_job:
                                # Job control
                                if is_infile_option(option_name):
                                    input_files.append(map_node_value)
                                    map_node_value = os.path.basename(
                                        map_node_value)

                                if is_outfile_option(option_name):
                                    map_node_value = os.path.basename(
                                        map_node_value)
                                    output_files.append(map_node_value)

                            cmd[pos] = KNIMEBatch._replaceOptionInCmd(
                                option_name, map_node_value, arg)

                            break
            pos = pos + 1

        # Append any unknown arguments so that those wil be simply passed to the
        # underlying script
        cmd.extend(other_args)

        ret_code = 0
        if not exec_cmd:
            # If -printknimecmd is true, send -printcmd to KNIME_workflow_wrapper
            # also. This will just print the command and exit. We capture the
            # stderr in a file and parse it to print the actual command.
            cmd.append("-printcmd")
            with open(os.path.join(self.tempdir, "cmdfile"), "w") as tmpfile:
                ret_code = subprocess.call(cmd, stderr=tmpfile)

        if sys.platform == 'win32':
            # KNIME-3939 Set SCHRODINGER_STU_TEST_ID so that knime.log will be
            # written in current directory
            os.environ['SCHRODINGER_STU_TEST_ID'] = "1"

        # Execute the command
        self.logger.info("Command to be executed:\n" + str(cmd))
        if host and launch_job:
            # Start KNIME under jobcontrol
            ret_code = self.launchKnime(host, cmd, input_files, output_files)
        else:
            # Start KNIME as normal command
            ret_code = subprocess.call(cmd)

        return ret_code

    #***************************************************************************
    def launchKnime(self, host, cmd, input_files, output_files):
        '''
        Launch KNIME_workflow_wrapper.py under jobcontrol
        While running under jobcontrol, do not specify absolute paths
        '''
        scriptlauncher = launcher.Launcher(script='KNIME_workflow_wrapper.py',
                                           prog='KNIME',
                                           jobname='KNIME_batch_job',
                                           wait=False,
                                           host=host,
                                           no_redirect=False)

        # $SCHRODINGER/run KNIME_workflow_wrapper.py workflow_path <args> ...
        # Get only the script args
        script_args = cmd[2:]
        scriptlauncher.addScriptArgs(script_args)

        #
        # Add input files
        #
        scriptlauncher.addInputFile(self.workflow)
        for infile in input_files:
            scriptlauncher.addInputFile(infile)

        #
        # Add output files
        #

        # Add destination workflow
        # Get the destination workflow if specified in command line from the
        # below option:
        # -destFile=<executed_workflow.knwf>
        for arg in cmd:
            if arg.startswith('-destFile='):
                dest_workflow = os.path.basename(arg.split('=')[1])
                scriptlauncher.addOutputFile(dest_workflow)
                break

        for outfile in output_files:
            scriptlauncher.addOutputFile(outfile)

        try:
            job = scriptlauncher.launch()
            ret_code = 0 if job else -1
        except (IOError, OSError, RuntimeError) as err:
            print(f'Launch failed: Error:{str(err)}')
            ret_code = -1

        return ret_code

    #***************************************************************************
    def redo_parseargs(self, args):

        self.logger.info("In redo_parseargs()")
        redo_parser = argparse.ArgumentParser()
        for option in self.option_map:
            redo_parser.add_argument(option, action="store", dest=option)

        redo_options, other_args = redo_parser.parse_known_args(args=args)

        self.logger.debug("Redo options: " + str(redo_options))
        self.logger.debug("Other args: " + str(other_args))

        # Populate the option_map with the values
        for option in self.option_map:
            self.option_map[option][1] = getattr(redo_options, option)

        self.logger.info("Option map:" + str(self.option_map))

        return redo_options, other_args

    #***************************************************************************
    def showGUI(self):
        from KNIME_template_gui import KnimeTemplateGUI

        self.logger.debug("In showGUI()")
        workflowGUI = KnimeTemplateGUI(self.workflow_parser, False, self.logger)
        workflowGUI.show()
        workflowGUI.raise_()
        # Start the event loop
        workflowGUI.exec()

    def cleanup(self):
        self.logger.debug(f"Deleting temporary directory '{self.tempdir}'")
        fileutils.force_rmtree(self.tempdir)


# KNIMEBatch class ends --
#*******************************************************************************


def run_knimepy(workflow, input_file):
    '''
    Execute a workflow via knime.py
    '''
    # Verify the workflow path
    workflow_abspath = validate_workflow_path(workflow)

    # If workflow is a zipfile, extract it to tempdir
    # knime.Wirkflow class only handles workflow dirs and not knwf files
    tempdir = tempfile.mkdtemp(prefix="KNIME_", dir=os.getcwd())
    processed_workflow_path = unzip_workflow(workflow_abspath, tempdir)

    # Prepare the pandas.DataFrame from input file
    molecules = []
    molecule_names = []
    indices = []
    for idx, st in enumerate(structure.StructureReader(input_file)):
        molecules.append(st.writeToString(structure.SD))
        molecule_names.append(st.title)
        indices.append(f'Row{idx}')

    input_table = pandas.DataFrame(
        {
            'Molecule': molecules,
            'Molecule name': molecule_names
        }, index=indices)

    # Execute the workflow
    import knime
    # Set knime.executable_path
    knime_bin_dir = hunt('knime')
    if sys.platform == "darwin":
        knime.executable_path = os.path.join(knime_bin_dir, "knime",
                                             "Knime.app", "Contents", "MacOS",
                                             "Knime")
    else:
        knime.executable_path = os.path.join(knime_bin_dir, "knime", "knime")
    with knime.Workflow(processed_workflow_path) as wkf:
        wkf.data_table_inputs[0] = input_table
        wkf.save_after_execution = True
        wkf.execute()

    # Create executed.knwf
    (parent_dir,
     workflow_name) = os.path.split(os.path.abspath(processed_workflow_path))
    shutil.make_archive('executed', 'zip', parent_dir, workflow_name)
    shutil.move('executed.zip', 'executed.knwf')

    fileutils.force_rmtree(tempdir)


def create_parser():
    # Parses the command line arguments
    parser = argparse.ArgumentParser(
        usage=USAGE,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=SCRIPT_DESCRIPTION,
        prog="KNIME_batch.py")

    parser.add_argument(
        "workflow",
        action="store",
        help="Path to the zipped workflow file or workflow directory")

    parser.add_argument(
        "-print_allowed",
        "-printallowed",
        action=PrintAllowedNodes,
        help="Print the allowed nodes that this script can detect in a workflow"
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "-print_usage",
        "-printusage",
        action="store_true",
        dest="printusage",
        help="Print the usage message with script specific options")

    group.add_argument(
        "-print_cmd",
        "-printcmd",
        action="store_true",
        dest="printcmd",
        help="Print the KNIME_batch.py's command to execute the workflow")

    group.add_argument("-print_knime_cmd",
                       "-printknimecmd",
                       action="store_true",
                       dest="printknimecmd",
                       help="Print the KNIME's batch execution command")

    group.add_argument("-stu",
                       action="store_true",
                       help="Print the command to be used in STU")

    group.add_argument("-run", action="store_true", help="Executes the command")

    group.add_argument("-GUI",
                       action="store_true",
                       dest="gui",
                       help="Displays a Python panel")

    parser.add_argument("-iknimepy",
                        help="Input file to run workflow using knime.py")

    parser.add_argument("-debug",
                        action="store_true",
                        help="Displays debug messages in the terminal")

    parser.add_argument("-keeptempfiles",
                        action="store_true",
                        help='''Do not delete the temporary directory if this
                        option is set. By default temporary files are deleted'''
                       )

    parser.add_argument(
        "-HOST",
        help='''Specify a host to run the node's backend, like: <host>:<cpus>.
        To launch the node's backend on this host, set PRIMARY_HOST in the
        Jobcontrol tab of node configuration. If -launchjob is also specified,
        then KNIME is launchd on the specified host under jobserver''')

    parser.add_argument(
        "-launchjob",
        action="store_true",
        help='''Launch KNIME as a job on the host specified in -HOST. To be used
        only when -run and -HOST are specified.''')

    # KNIME-5599 If this option is specified, just ignore. This is required to
    # specify some default option value in KNIME LiveDesign protocols
    parser.add_argument("-dummy", action="store_true", help=argparse.SUPPRESS)

    return parser


#*******************************************************************************
class PrintAllowedNodes(argparse.Action):
    '''
    This class is an inherited Custom action class to print the allowed nodes
    and the node labels that can be identified by
    KNIME_workflow_wrapper.parseWorkflow() method. Without this custom class,
    argparse expects -printusage, -printknimecmd, or -run even when -printallowed is
    specified."
    '''

    def __init__(self, *args, **kwargs):
        super(PrintAllowedNodes, self).__init__(*args, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        print_allowed()
        sys.exit(0)


def print_allowed():
    print('''
==========================================================================================
Note: Entries marked as "Maestro only" should be used within Maestro

------------------------------------------------------------------------------------------
Node Name              Node Label           Option        Description
------------------------------------------------------------------------------------------
Schrodinger READERS
------------------------------------------------------------------------------------------
Molecule Reader        %input% | %file%     -istruct      Read the first file specified in
                                                          the node.
                       %selected entries%                 Read the entries that are
                                                          selected in the Project Table
                                                          (Maestro only).
                       %included entries%                 Read the entries that are
                                                          included in the Workspace
                                                          (Maestro only).
Alignment Reader       %input% | %file%     -ialign       Read the file specified in the
                                                          node.
Sequence Reader        %input% | %file%     -iseq         Read the file specified in the
                                                          node.
Smiles Reader          %input% | %file%     -ismi         Read the first file specified in
                                                          the node.
PDB Reader             %input% | %file%     -ipdb         Read the file specified in the
                                                          node.
Glide Grid Reader      %input% | %file%     -igrd         Read the first file specified in
                                                          the node.
Hypothesis Reader      %input% | %file%     -ihyp         Read the first file specified in
                                                          the node.
Canvas Fingerprint     %input% | %file%     -ifp          Read the file specified in the
Reader                                                    node.
QSAR Model Reader      %input% | %file%     -iqzip        Read the file specified in the
                                                          node.
Desmond CMS Reader     %input% | %file%     -icms         Read the first file specified in
                                                          the node.
FEP+ Reader            %input% | %file%     -ifmp         Read the first file specified in
                                                          the node.
------------------------------------------------------------------------------------------
KNIME READERS
------------------------------------------------------------------------------------------
CSV Reader             %input% | %file%     -itxt         Read the file specified in the
                                                          node.
SDF Reader             %input% | %file%     -isdf         Read the first file specified in
                                                          the node.
Mol2 Reader            %input% | %file%     -imol2        Read the file specified in the
                                                          node.
Excel Reader           %input% | %file%     -ixls         Read the file specified in the
                                                          node.
Model Reader           %input% | %file%     -ikmodel      Read the file specified in the
                                                          node.
Table Reader           %input% | %file%     -itable       Read the file specified in the
                                                          node.
List Files             %input% | %file%     -ilistfiles   Read the location specified in
                                                          the node.

------------------------------------------------------------------------------------------
Schrodinger WRITERS
------------------------------------------------------------------------------------------
Molecule Writer        %file%               -ostruct      Write the structures to the
                                                          specified file.
                       %new group%                        Add the output structures as a
                                                          new group to the Project Table
                                                          (Maestro only).
Alignment Writer       %file%               -oalign       Write the output to specified
                                                          file.
Sequence Writer        %file%               -oseq         Write the output to specified
                                                          file.
Glide Grid Writer      %file%               -ogrd         Write the output to specified
                                                          file.
Hypothesis Writer      %file%               -ohyp         Write the output to specified
                                                          directory.
Canvas Fingerprint     %file%               -ofp          Write the output to specified
Writer                                                    file.
QSAR Model Writer      %file%               -oqzip        Write the output to specified
                                                          file.

------------------------------------------------------------------------------------------
KNIME WRITERS
------------------------------------------------------------------------------------------
CSV Writer             %file%               -otxt         Write the output to specified
                                                          file.
                       %maestro message%                  Display a message in an
                                                          information box. The message is
                                                          the output of the CSV Writer
                                                          node (Maestro only).
                       %maestro commands%                 Execute Maestro commands on
                                                          completion of the workflow and
                                                          incorporation of the results.
                                                          The commands are in the first
                                                          column of the CSV Writer output
                                                          (Maestro only).
SDF Writer             %file%               -osdf         Write the output to specified
                                                          file.
MOL2 Writer            %file%               -omol2        Write the output to specified
                                                          file.
Excel Writer           %file%               -oxls         Write the output to specified
                                                          file.
Table Writer           %file%               -otable       Write the output to specified
                                                          file.

------------------------------------------------------------------------------------------
QUICK FORM NODES (both legacy and new ones)
------------------------------------------------------------------------------------------
Integer Configuration, %input% or           -i<var_name>  Add a text box control to the
Double Configuration,  %guisetting%                       GUI panel and assign the value
String Configuration,                                     that is set in the text box.
List Box Configuration

Integer Input,         %input% or           -i<var_name>  Add a text box control to the
Double Input,          %guisetting%                       GUI panel and assign the value
String Input                                              that is set in the text box.

Boolean Input          %input% or           -i<var_name>  Add a checkbox control to the
Boolean Configuration, %guisetting%                       GUI panel.

String Radio Buttons   %guisettingcombo%    -i<var_name>  Add a set of radio buttons to
                                                          the GUI panel presenting
                                                          alternatives to be selected.

String Radio Buttons   %guisettingradio%    -i<var_name>  Add a combo box (option menu) to
                                                          the GUI panel presenting
                                                          alternatives to be selected.

File Upload            %input% or           -i<var_name>  Add a text box with a "Browse"
                       %guisetting%                       button to select a file.
==========================================================================================
''')


#*******************************************************************************

if __name__ == '__main__':

    # Parse the command line args
    parser = create_parser()
    (options, redo_args) = parser.parse_known_args(args=sys.argv[1:])

    ret_code = 0
    knimepy_input_file = options.iknimepy
    if knimepy_input_file:
        # Execute workflow via knime.py
        try:
            run_knimepy(options.workflow, knimepy_input_file)
        except ValueError as ve:
            print(ve)
            ret_code = 1
        sys.exit(ret_code)

    # If control comes here, then execute workflow via KNIME_batch
    # Create KNIMEBatch instance
    try:
        knime_batch = KNIMEBatch(options.workflow, options.debug)
    except ValueError as ve:
        print(ve)
        exit(1)

    if options.printusage:
        print(knime_batch.getHelpMessage(PrintType.USAGE))
    elif options.printcmd:
        print(knime_batch.getHelpMessage(PrintType.COMMAND))
    elif options.stu:
        print(knime_batch.getHelpMessage(PrintType.STU_COMMAND))
    elif options.run:
        ret_code = knime_batch.execute(redo_args, launch_job=options.launchjob)
    elif options.printknimecmd:
        ret_code = knime_batch.execute(redo_args, exec_cmd=False)
    elif options.gui:
        knime_batch.showGUI()

    if not options.keeptempfiles:
        knime_batch.cleanup()

    sys.exit(ret_code)
