import argparse
import glob
import logging as log
import os
from pathlib import Path
import re
import subprocess
import sys
import tempfile
import zipfile

# KNIME update site
KNIME_UPDATESITE = "http://update.knime.org/analytics-platform/"
# Trusted Community Contributions update site
TRUSTED_COMMUNITIES_UPDATESITE = "http://update.knime.org/community-contributions/trusted/"

SCRIPT_DESCRIPTION = '''
Utility to automatically install KNIME features and Schrodinger KNIME extensions
in KNIME Analytics Platform
'''

SCRIPT_EPILOG = '''
Examples:
#1 
To install Schrodinger extensions in an existing KNIME AP:

"$SCHRODINGER/run" -FROM knime KNIME_install.py /path/to/KNIME
-schrod_updatesite /path/to/SchrodingerKNIMEUpdateSite_<version>.zip

#2
To install third-party extensions in an existing KNIME AP:

"$SCHRODINGER/run" -FROM knime KNIME_install.py /path/to/KNIME
-updatesites 4.6 -features_list /path/to/features.txt

where, features.txt should contain one feature per line, as below:
org.rdkit.knime.feature.feature.group
org.rdkit.knime.wizards.feature.feature.group

#3
To install Schrodinger and a couple of other extensions in KNIME AP:

"$SCHRODINGER/run" -FROM knime KNIME_install.py /path/to/KNIME 
-schrod_updatesite /path/to/SchrodingerKNIMEUpdateSite_<version>.zip
-updatesites 4.6 -features_list /path/to/features.txt
'''


def parse_args():
    # Parses the command line arguments
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=SCRIPT_DESCRIPTION,
        epilog=SCRIPT_EPILOG)

    parser.add_argument("knime_directory",
                        help='''
                        Path to KNIME AP directory. If a regex is
                        specified, the path that matches first will be taken. On
                        MacOS, this is the directory containing Knime.app.
                        ''')

    parser.add_argument("-schrod_updatesite",
                        help='''
                        Specify the Schrodinger zipped update site file path.
                        ''')
    parser.add_argument("-updatesites",
                        help='''
                        Specify the list of comma separated update sites.
                        The update sites can be the actual update site web url,
                        a local file path to a zipped update site or a simple
                        version number like '3.1'.
                        If a version number is specified, then the KNIME and
                        Trusted Community Contributions update site urls of that
                        particular version will be added automatically.
                        If this option is not specified, and if '-features_list'
                        is specified, then the KNIME and Trusted Community
                        Contributions update site urls of the version
                        corresponding to the KNIME directory specified in
                        '-knime_directory' will be used automatically.
                        ''')
    parser.add_argument("-features_list",
                        help='''
                        File with a list of features to be installed (one per
                        line). If -updatesites is specified, then the features
                        will be installed from these update sites else the KNIME
                        update site for the current version of KNIME AP is used.
                        ''')
    parser.add_argument("-install_basechem",
                        dest="basechem_version",
                        help='''
                        Install 'KNIME Base Chemistry Types & Nodes' feature by
                        specifying a version number like '4.6'. The feature will
                        be installed from the KNIME update site.
                        If this option is not specified, then this feature will
                        not be installed at all. In this case make sure that it
                        is already installed before installing Schrodinger
                        extensions.
                        ''')
    parser.add_argument("-configuration",
                        help='''
                        Configuration location to install the plugins.
                        If the plugins needs to be installed under
                        /tmp/myplugins, then specify the configuration location
                        as /tmp/myplugins/configuration
                        ''')
    parser.add_argument("-verbose",
                        action="store_true",
                        help="Prints debug level messages")
    # Not used anymore
    parser.add_argument("-keeplogs",
                        action="store_true",
                        help=argparse.SUPPRESS)
    # -stats is used for internal purpose. Not intended to be used by customers
    parser.add_argument("-stats", action="store_true", help=argparse.SUPPRESS)

    args = parser.parse_args()

    return args


def validate_args(args):
    """
    Validate the command line args

    :param args: argparse namespace object
    :type args: argparse.Namespace

    :return: True if valid else False
    :rtype: bool
    """
    status = True
    if not (args.schrod_updatesite or args.features_list or
            args.basechem_version or args.stats):
        print(
            "At least one of these options must be specified: schrod_updatesite|install_basechem|features_list"
        )
        status = False

    if args.updatesites and not args.features_list:
        print(
            "If '-updatesites' is specified, '-features_list' must be specified"
        )
        status = False

    return status


def create_logger(log_level=log.WARNING):
    """
    Create a `logging.Logger` instance and return the same.

    :param log_level: Log level
    :type log_level: int

    :return: `logging.Logger` instance
    :rtype: logging.Logger
    """
    logger = log.getLogger("")
    logger.setLevel(log_level)

    handler = log.StreamHandler(sys.stdout)
    handler.setFormatter(log.Formatter("%(levelname)s::%(message)s"))
    logger.addHandler(handler)

    return logger


def write_file_lengths(directory):
    """
    Writes a file with <file_length> <file_name> for each file under `directory`
    and returns the file path.

    :param directory: Path to a directory for which file lengths should be
                      written
    :type directory: str

    :return: Path to the file containing <file_length> <file_name>
    :rtype: str
    """
    dir_p = Path(directory)
    files_list = []
    for filepath in dir_p.rglob('*'):
        relpath = str(filepath.relative_to(dir_p))
        files_list.append([str(len(relpath)), relpath])

    (handle, outfile) = tempfile.mkstemp(suffix='.txt',
                                         prefix='filelength_',
                                         dir=Path.cwd())
    with open(outfile, 'w') as fd:
        for line in sorted(files_list, reverse=True, key=lambda i: int(i[0])):
            fd.write(' '.join(line) + os.linesep)

    return outfile


def get_matched_path(path_regex):
    """
    Return the first matched path for the path regex given. Return None if
    nothing matched

    :param path_regex: A path which may contain regex
    :type path_regex: str

    :return: First matched path or None
    :rtype: str or None
    """
    first_matched_path = None
    matched_paths_arr = glob.glob(path_regex)
    if len(matched_paths_arr) > 0:
        first_matched_path = os.path.abspath(matched_paths_arr[0])

    return first_matched_path


def read_features_file(features_file):
    """
    Read the features from `features_file` into a list and return the same.
    The feature file should contain one feature per line. Lines starting with #
    are comments. Comments can also appear in a line after a feature followed by
    a space. Empty lines and comments are ignored.

    Example file contents:
    # RDKit features
    org.rdkit.knime.feature.feature.group
    org.rdkit.knime.wizards.feature.feature.group

    com.schrodinger.knime.feature.nodes.feature.group # Schrodinger feature
    <eof>

    :param features_file: Path to the features file
    :type features_file: str

    :return: List of features read from the file
    :rtype: list
    """
    features_list = []
    with open(features_file, 'r') as fd:
        for line in fd:
            line = line.strip()
            # Ignore comments and empty lines
            if not line or line.startswith('#'):
                continue
            feature = line.split(' ', 1)[0]
            feature = feature.strip()
            features_list.append(feature)

    return features_list


class ExtensionsInstaller:
    """
    Install additional extensions to a KNIME Analytics Paltform
    """
    SCHRODINGER_FEATURE = 'com.schrodinger.knime.feature.nodes.feature.group'
    BASECHEM_FEATURE = 'org.knime.features.chem.types.feature.group'
    SCHRODINGER_PLUGINS = ('com.schrodinger.knime.base',
                           'com.schrodinger.knime.nodes',
                           'com.schrodinger.knime.nodes.deprecated',
                           'com.schrodinger.knime.dynamicnodes',
                           'com.schrodinger.knime.2drenderers')

    def __init__(self, knime_dir, config_dir=None, logger=None):
        """
        :param knime_dir: Path to the KNIME installation directory aka 'target
         KNIME'
        :type knime_dir: str

        :param config_dir: Path to the user configuration location. If not
                           passed, the default config location under knime_dir
                           will be used
        :type config_dir: str or None

        :param logger: logging.Logger instance. If not passed, a new loggger
                       insatnce is created.
        :type logger: `logging.Logger` or None
        """
        self.knime_dir = Path(knime_dir).resolve(strict=True)
        if sys.platform == 'darwin':
            self.knime_exec = self.knime_dir.joinpath(
                'Knime.app', 'Contents', 'MacOS', 'knime').resolve(strict=True)
            self.eclipse_dir = self.knime_dir.joinpath(
                'Knime.app', 'Contents', 'Eclispe').resolve(strict=True)
        else:
            self.knime_exec = self.knime_dir.joinpath('knime').resolve(
                strict=True)
            self.eclipse_dir = self.knime_dir.resolve(strict=True)

        self.base_cmd = (str(self.knime_exec), '-consoleLog', '-nosplash',
                         '-application', 'org.eclipse.equinox.p2.director')
        if config_dir:
            self.base_cmd += ('-configuration', config_dir)

        self.logger = logger if logger else create_logger()
        try:
            self.knime_version = self.getKNIMEVersion()
        except (FileNotFoundError, ValueError) as error:
            self.knime_version = None
            self.logger.warning(error)

    def setLogLevel(self, log_level):
        """
        Set logger's log level

        :param log_level: Log level
        :type log_level: int
        """
        self.logger.setLevel(log_level)

    def getKNIMEVersion(self):
        """
        Get the KNIME version (major.minor) from about.ini file from
        org.knime.product plugin under target KNIME

        Line to be parsed from about.ini:
        aboutText=KNIME Version: 4.6.1

        :return: KNIME version in format major.minor
        :rtype: str

        :raises FileNotFoundError: If about.ini file not found in target KNIME
        :raises ValueError: If KNIME version could not be parsed from about.ini
        """
        # plugins path
        plugin_path = self.eclipse_dir.joinpath('plugins')

        # Get about.ini file
        about_ini_file_arr = list(
            plugin_path.glob("org.knime.product_*/about.ini"))
        if not about_ini_file_arr:
            raise FileNotFoundError(
                f"No about.ini file found in the KNIME directory: '{self.knime_dir}'"
            )

        about_ini_file = about_ini_file_arr[0]
        version = None
        with open(about_ini_file, "r") as ini_obj:
            for line in ini_obj.readlines():
                line = line.strip()
                match = re.match(
                    r'^aboutText=KNIME Version: (\d+\.\d+)\.\d+.*$', line)
                if match:
                    version = match.group(1)
                    break

        if not version:
            raise ValueError(
                f"Could not parse KNIME version from '{about_ini_file}'")

        return version

    def getSchrodExtVersion(self, schrod_updatesite):
        """
        Get the Schrodinger extensions version from the zipped update site
        `schrod_updatesite`

        :param schrod_updatesite: Schrodinger zipped update site filepath
        :type schrod_updatesite: str

        :return: Schrodinger extensions version or None if version could not be
                 parsed
        :rtype: str or None
        """
        version = None
        if not Path(schrod_updatesite).exists():
            self.logger.error(f'File not found: {schrod_updatesite}')
            return version

        with zipfile.ZipFile(schrod_updatesite, "r") as zf:
            for name in zf.namelist():
                match = re.search(r".*com\.schrodinger\.knime\.base_(.+?)\.jar",
                                  name)
                if match:
                    version = match.group(1)
                    break

        return version

    def validateSchrodExtInstallation(self, schrod_ext_version):
        """
        Validate the Schrodinger extensions installed in target KNIME. Check if
        the version installed is same as `schrod_ext_version`. This is done by
        checking if bundles.info file under target KNIME has all the
        Schrodinger plugins matching the `schrod_ext_version`

        Sample line in bundles.info file:
        com.schrodinger.knime.base,23.2.26.v202301290408,plugins/com.schrodinger.knime.base_23.2.26.v202301290408/,4,false

        :param schrod_ext_version: Schrodinger extensions version like, 23.2.26.v202301290408
        :return: True if Schrodinger extensions for the given version are
                 installed else False.
        :rtype: bool
        """
        bundles_info_file = self.eclipse_dir.joinpath(
            'configuration', 'org.eclipse.equinox.simpleconfigurator',
            'bundles.info')

        if not bundles_info_file.exists():
            self.logger.error(f"File not found: {bundles_info_file}")
            return False

        # Read bundles.info in a list
        with open(bundles_info_file, "r") as fd:
            bundles_info_list = fd.readlines()

        for schrod_plugin in self.SCHRODINGER_PLUGINS:
            schrod_plugin_regex = re.compile(
                f"^.+?{schrod_plugin}_{schrod_ext_version}.+?$")

            if not list(filter(schrod_plugin_regex.match, bundles_info_list)):
                return False

        return True

    def getInstalledFeatures(self):
        """
        Get the output of Eclipse p2 director application's
        '-listInstalledRoots' option that contains the text with all installed
        features in the target KNIME

        :return: Installed features
        :rtype: str

        :raises subprocess.CalledProcessError: If the command to get the
            installed features fails
        """
        self.logger.debug("Getting the list of already installed features")
        cmd = list(self.base_cmd)
        cmd.append('-listInstalledRoots')
        cproc = subprocess.run(cmd, capture_output=True, text=True, check=True)
        # stdout of the process contains all the installed features
        return cproc.stdout

    def getFeaturesToUninstall(self, features_to_install):
        """
        Get the list of features to uninstall. If a feature in
        `features_to_install` list is already installed in target KNIME, then it
        needs to be uninstalled

        :param features_to_install: List of features to install
        :type features_to_install: list 

        :return: List of features to uninstall
        :rtype: list

        :raises RuntimeError: If getting the features to uninstall fails
        """
        self.logger.debug("Getting the list of features to be uninstalled")
        features_to_uninstall = []
        try:
            installed_features = self.getInstalledFeatures()
        except subprocess.CalledProcessError as error:
            self.logger.warning(
                "Failed to get the installed features from target KNIME")
            self.logger.debug(error.stdout)
            self.logger.debug(error.stderr)
            raise RuntimeError("Failed to get the features to uninstall")

        for feature in features_to_install:
            if feature in installed_features:
                features_to_uninstall.append(feature)

        return features_to_uninstall

    def validVersion(self, version):
        """
        Check if the `version` is in valid format: major.minor

        :param version: KNIME Version
        :type version: str

        :return: True if `version` is in valid format else False
        :rtype: bool
        """
        match = re.match(r"\d+\.\d+$", version)
        return bool(match)

    def prepareLocalRepositoryURL(self, path):
        """
        Prepare the `path` (either a zip file or a directory) to be used as
        local repository URL

        :param path: Local path to be prepared
        :type path: str

        :return: Prepared URL
        :rtype: str

        :raises FileNotFoundError: If `path` does not exist
        """
        pathobj = Path(path)
        uri = pathobj.resolve(strict=True).as_uri()
        if pathobj.is_file():
            return f'jar:{uri}!/'
        else:
            return uri

    def prepareRepositories(self, repositories):
        """
        Prepare the list of `repositories` passed to be used as repository URLs.
        The `repositories` can be the actual remote update site url, a local file
        path to a zipped update site, a local directory containing the update
        site or a simple version number like '4.6'.
        If a remote url is specified, then it will be passed as it is.
        If a local path is specified (file or directory), it will be converted
        to a repository url by calling `prepareLocalRepositoryURL()`
        If a version number is specified, then it will be converted to KNIME and
        Trusted Community Contributions update site urls of that particular
        version

        :param repositories: List of repositories
        :type repositories: list

        :return: Prepared list of repositories
        :rtype: list
        """
        prepared_repositories = []
        for repo in repositories:
            if self.validVersion(repo):
                # It is a version number like '4.6'. So add the Knime and
                # Trusted community repositories for this version
                prepared_repositories.append(f'{KNIME_UPDATESITE}{repo}')
                prepared_repositories.append(
                    f'{TRUSTED_COMMUNITIES_UPDATESITE}{repo}')
            elif Path(repo).exists():
                # Local repository path
                prepared_repositories.append(
                    self.prepareLocalRepositoryURL(repo))
            else:
                # Should be an update site url. Just append it
                prepared_repositories.append(repo)

        return prepared_repositories

    def installFeatures(self, features, repositories=None):
        """
        Install the list of `features` from the list of `repositories` on the
        target KNIME

        :param features: List of features
        :type features: list

        :param repositories: Optional list of repositories
        :type: list or None

        :return: True if installation success else False
        :rtype: bool
        """
        self.logger.debug("Installing features ...")
        if not features:
            self.logger.error("Features list should be non-empty")
            return False

        # Prepare repositories
        if not repositories:
            self.logger.info("No repositories specified")
            if not self.knime_version:
                self.logger.error(
                    "Could not get current KNIME version to contact the KNIME update site"
                )
                return False

            self.logger.info(
                f"Installing features from KNIME update site for current KNIME version {self.knime_version}"
            )
            repositories = [self.knime_version]

        prepared_repositories = self.prepareRepositories(repositories)

        # If a feature to be installed is already installed, then installation
        # will fail for some features, as more than one version of such feature
        # or its dependencies cannot co-exist in an installation profile. So
        # uninstall such features before installing.
        features_to_uninstall = []
        try:
            features_to_uninstall = self.getFeaturesToUninstall(features)
        except RuntimeError as error:
            self.logger.warning(error)
            self.logger.warning(
                "Proceeding to install the existing features, which may fail")

        if self.logger.isEnabledFor(log.DEBUG):
            self.logger.debug("List of features to be installed:")
            for feature in features:
                self.logger.debug(feature)

        cmd = list(self.base_cmd)
        cmd.append('-repository')
        cmd.append(','.join(prepared_repositories))
        cmd.append('-installIU')
        cmd.append(','.join(features))
        if features_to_uninstall:
            cmd.append('-uninstallIU')
            cmd.append(','.join(features_to_uninstall))

        cmd_line = ' '.join(cmd)
        self.logger.info(f"Command to be executed: '{cmd_line}'")

        cproc = subprocess.run(cmd,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               text=True)
        if cproc.returncode != 0:
            self.logger.critical(cproc.stdout)
            return False

        self.logger.debug(cproc.stdout)
        return True

    def installSchrodExtensions(self, updatesite):
        """
        Install the Schrodinger extensions from the `updatesite` on target KNIME

        :param updatesite: Zipped update site containing Schrodinger extensions
        :type updatesite: str

        :return: True if installation success else False
        :rtype: bool
        """
        self.logger.debug("Installing Schrodinger KNIME extensions")
        schrod_ext_version = self.getSchrodExtVersion(updatesite)
        if not schrod_ext_version:
            self.logger.error(
                f"Error in getting Schrodinger extensions version from '{updatesite}'"
            )
            return False

        self.logger.info(
            f"Schrodinger KNIME extensions version to be installed is: {schrod_ext_version}"
        )

        updatesites = [updatesite]
        # Add KNIME update site also, as Schrodinger extensions has some
        # dependencies on that
        if self.knime_version:
            updatesites.append(self.knime_version)
        else:
            self.logger.warning(
                "Could not add KNIME update site while installing Schrodinger extensions"
            )

        if not self.installFeatures([self.SCHRODINGER_FEATURE], updatesites):
            return False

        if not self.validateSchrodExtInstallation(schrod_ext_version):
            self.logger.error(
                "Validation failed as at least one of the Schrodinger plugins could not be found in bundles.info file"
            )
            return False

        return True

    def installBaseChem(self, version):
        """
        Install KNIME's 'Base Chemistry Types & Nodes' feature for the given
        `version`

        :param version: Base chemistry types feature version to be installed
        :type version: str

        :return: True if installation success else False
        :rtype: bool
        """
        self.logger.debug("Installing 'Base Chemistry Types & Nodes'")
        if not self.validVersion(version):
            self.logger.error(
                f"Invalid version passed: {version}. Version should be in format: <major>.<minor>"
            )
            return False

        return self.installFeatures([self.BASECHEM_FEATURE], [version])


def main():
    """
    Parse the command line arguments and installs the specified extensions in
    KNIME Analytics Platform

    :return: True on success else False
    :rtype: bool
    """
    # Parse the command line args
    args = parse_args()

    # Validate args
    valid = validate_args(args)
    if not valid:
        return False

    logger = create_logger()
    if args.verbose:
        logger.setLevel(log.DEBUG)

    # Check existence of knime_directory
    knime_dir = get_matched_path(args.knime_directory)
    if knime_dir:
        logger.debug(
            f"KNIME Analytics Platform directory matched is '{knime_dir}'")
    else:
        logger.error(
            f"No KNIME Analytics Platform directory matched for path '{args.knime_directory}'"
        )
        return False

    # Check existence of features_list file
    features_list = []
    if args.features_list:
        features_file = get_matched_path(args.features_list)
        if features_file:
            logger.debug(f"Features file matched is '{features_file}'")
            # Read features file into a list
            features_list = read_features_file(features_file)
            if not features_list:
                logger.error(
                    f"Features file '{features_file}' does not contain any feature"
                )
                return False
        else:
            logger.error(
                f"No features file matched for path '{args.features_list}'")
            return False

    status = True
    if args.schrod_updatesite or args.features_list or args.basechem_version:
        installer = ExtensionsInstaller(knime_dir,
                                        config_dir=args.configuration,
                                        logger=logger)

        if args.basechem_version:
            status = installer.installBaseChem(args.basechem_version)

        if features_list:
            updatesites_list = []
            if args.updatesites:
                # Build updatesites list from comma separated string
                updatesites_list = args.updatesites.split(',')

            status = installer.installFeatures(features_list, updatesites_list)

        if args.schrod_updatesite:
            status = installer.installSchrodExtensions(args.schrod_updatesite)

        if status:
            print("INSTALLATION SUCCESS")
        else:
            print("INSTALLATION FAILED")

    if args.stats:
        outfile = write_file_lengths(knime_dir)
        print(f"File path lengths written to: {outfile}")

    return status


if __name__ == '__main__':
    ret_status = main()
    exit_code = 0 if ret_status else 1
    sys.exit(exit_code)
