# KNIME_workflow_wrapper.py
# version 0.9
# Contributors: Jean-Christophe Mozziconacci, Ravikiran Kuppuraj,
#   Matvey Adzhigirey, Blaine Bell
"""
KNIME_workflow_wrapper

This script runs a KNIME workflow from Maestro.

Copyright (c) 2011, Schrodinger
"""

import argparse
import glob
import logging as log
import os
import random
import re
import shlex
import shutil
import string
import sys
import tarfile
import tempfile
import time
import xml.dom.minidom
from zipfile import ZipFile

import schrodinger
import schrodinger.utils.maestro_scripts_path as maescripts
from schrodinger import structure
from schrodinger.job import jobhandler
from schrodinger.job import launcher
from schrodinger.utils import fileutils
from schrodinger.utils import mmutil
from schrodinger.utils import subprocess

maestro = schrodinger.get_maestro()

# Global node names
# Schrodinger Readers
MOLECULE_READER = "com.schrodinger.knime.node.maemultifilereader.MAEMultiFileReaderNodeFactory"
ALIGNMENT_READER = "com.schrodinger.knime.node.alignreader.AlignmentReaderNodeFactory"
SEQUENCE_READER = "com.schrodinger.knime.node.seqreader.SequenceReaderNodeFactory"
SMILES_READER = "com.schrodinger.knime.node.smilesreader.SmilesFileReaderNodeFactory"
PDB_READER = "com.schrodinger.knime.node.pdbreader.PDBFileReaderNodeFactory"
FP_READER = "com.schrodinger.knime.node.fpreader.FingerprintReaderNodeFactory"
GRID_READER = "com.schrodinger.knime.node.glidegridmultireader.GlideGridMultiReaderNodeFactory"
HYP_READER = "com.schrodinger.knime.node.hypothesisreader.HypothesisReaderNodeFactory"
QSAR_READER = "com.schrodinger.knime.node.singlefilereader.QSARReaderFactory"
CMS_READER = "com.schrodinger.knime.node.desmondcmsreader.DesmondCMSReaderNodeFactory"
FEP_READER = "com.schrodinger.knime.node.fepreader.FEPReaderNodeFactory"
LD_INPUT = "com.schrodinger.knime.node.ldinput.LDInputNodeFactory"

# KNIME Readers
CSV_READER = "org.knime.base.node.io.filehandling.csv.reader.CSVTableReaderNodeFactory"
SDF_READER = "org.knime.chem.base.node.io.sdf.SDFReader3NodeFactory"
MOL2_READER = "org.knime.chem.base.node.io.mol2.Mol2ReaderFactory"
EXCEL_READER = "org.knime.ext.poi3.node.io.filehandling.excel.reader.ExcelTableReaderNodeFactory"
KNIME_MODEL_READER = "org.knime.base.node.io.filehandling.model.reader.ModelReaderNodeFactory"
KNIME_LIST_FILES = "org.knime.filehandling.utility.nodes.listpaths.ListFilesAndFoldersNodeFactory"
KNIME_TABLE_READER = "org.knime.base.node.io.filehandling.table.reader.KnimeTableReaderNodeFactory"

# Deprecated Readers after 4.4
CSV_READER_1 = "org.knime.base.node.io.csvreader.CSVReaderNodeFactory"
EXCEL_READER_1 = "org.knime.ext.poi2.node.read4.XLSReaderNodeFactory"
KNIME_MODEL_READER_1 = "org.knime.base.node.io.predictor.PredictorReaderNodeFactory"
KNIME_LIST_FILES_1 = "org.knime.base.node.io.listfiles2.ListFilesNodeFactory"
KNIME_TABLE_READER_1 = "org.knime.base.node.io.table.read.ReadTableNodeFactory"

ALLOWED_INPUT_NODES = [
    MOLECULE_READER, ALIGNMENT_READER, SEQUENCE_READER, CSV_READER, SDF_READER,
    MOL2_READER, SMILES_READER, PDB_READER, EXCEL_READER, FP_READER,
    GRID_READER, HYP_READER, QSAR_READER, CMS_READER, FEP_READER, LD_INPUT,
    KNIME_MODEL_READER, KNIME_LIST_FILES, KNIME_TABLE_READER
]

DEPRECATED_INPUT_NODES = [
    CSV_READER_1, EXCEL_READER_1, KNIME_MODEL_READER_1, KNIME_LIST_FILES_1,
    KNIME_TABLE_READER_1
]

# Nodes that can take input from LiveReport
LIVEREPORT_INPUT_NODES = [SDF_READER, CSV_READER, CSV_READER_1]

# Schrodinger Writers
MOLECULE_WRITER = "com.schrodinger.knime.node.maefilewriter.MAEFileWriterNodeFactory"
ALIGNMENT_WRITER = "com.schrodinger.knime.node.alignwriter.AlignmentWriterNodeFactory"
SEQUENCE_WRITER = "com.schrodinger.knime.node.seqwriter.SequenceWriterNodeFactory"
FP_WRITER = "com.schrodinger.knime.node.fpwriter.FingerprintWriterNodeFactory"
GRID_WRITER = "com.schrodinger.knime.node.glidegridwriter.GlideGridWriterNodeFactory"
HYP_WRITER = "com.schrodinger.knime.node.hypothesiswriter.HypothesisWriterNodeFactory"
QSAR_WRITER = "com.schrodinger.knime.node.qsarmodelwriter.QSARModelWriterNodeFactory"
# KNIME Writers
CSV_WRITER = "org.knime.base.node.io.filehandling.csv.writer.CSVWriter2NodeFactory"
SDF_WRITER = "org.knime.chem.base.node.io.sdfwriter2.SDFWriterNodeFactory"
MOL2_WRITER = "org.knime.chem.base.node.io.mol2.Mol2WriterNodeFactory"
EXCEL_WRITER = "org.knime.ext.poi3.node.io.filehandling.excel.writer.ExcelTableWriterNodeFactory"
KNIME_TABLE_WRITER = "org.knime.base.node.io.filehandling.table.writer.TableWriterNodeFactory"

# Deprecated Writers after 4.1.3
CSV_WRITER_1 = "org.knime.base.node.io.csvwriter.CSVWriterNodeFactory"
EXCEL_WRITER_1 = "org.knime.ext.poi2.node.write3.XLSWriter2NodeFactory"
KNIME_TABLE_WRITER_1 = "org.knime.base.node.io.table.write.WriteTableNodeFactory"

ALLOWED_OUTPUT_NODES = [
    MOLECULE_WRITER, ALIGNMENT_WRITER, SEQUENCE_WRITER, CSV_WRITER, SDF_WRITER,
    MOL2_WRITER, EXCEL_WRITER, FP_WRITER, GRID_WRITER, HYP_WRITER, QSAR_WRITER,
    KNIME_TABLE_WRITER
]

DEPRECATED_OUTPUT_NODES = [CSV_WRITER_1, EXCEL_WRITER_1, KNIME_TABLE_WRITER_1]

# Global Quick Form nodes
INTEGER_CONFIG = "org.knime.js.base.node.configuration.input.integer.IntegerDialogNodeFactory"
DOUBLE_CONFIG = "org.knime.js.base.node.configuration.input.dbl.DoubleDialogNodeFactory"
STRING_CONFIG = "org.knime.js.base.node.configuration.input.string.StringDialogNodeFactory"
BOOLEAN_CONFIG = "org.knime.js.base.node.configuration.input.bool.BooleanDialogNodeFactory"
LISTBOX_CONFIG = "org.knime.js.base.node.configuration.input.listbox.ListBoxDialogNodeFactory"

INTEGER_INPUT = "org.knime.js.base.node.quickform.input.integer.IntInputQuickFormNodeFactory"
DOUBLE_INPUT = "org.knime.js.base.node.quickform.input.dbl.DoubleInputQuickFormNodeFactory"
STRING_INPUT = "org.knime.js.base.node.quickform.input.string.StringInputQuickFormNodeFactory"
BOOLEAN_INPUT = "org.knime.js.base.node.quickform.input.bool.BooleanInputQuickFormNodeFactory"
FILE_UPLOAD = "org.knime.js.base.node.quickform.input.fileupload.FileUploadQuickFormNodeFactory"
RADIO_BUTTONS = "org.knime.js.base.node.quickform.selection.single.SingleSelectionQuickFormNodeFactory"

INTEGER_INPUT_LEGACY = "org.knime.quickform.nodes.in.integer.IntInputQuickFormInNodeFactory"
DOUBLE_INPUT_LEGACY = "org.knime.quickform.nodes.in.dbl.DoubleInputQuickFormInNodeFactory"
STRING_INPUT_LEGACY = "org.knime.quickform.nodes.in.string.StringInputQuickFormInNodeFactory"
BOOLEAN_INPUT_LEGACY = "org.knime.quickform.nodes.in.bool.BooleanInputQuickFormInNodeFactory"
FILE_UPLOAD_LEGACY = "org.knime.quickform.nodes.in.fileupload.FileUploadQuickFormInNodeFactory"
RADIO_BUTTONS_LEGACY = "org.knime.quickform.nodes.in.stringselection.StringSelectionInputQuickFormInNodeFactory"

ALLOWED_QF_NODES = [
    INTEGER_CONFIG, DOUBLE_CONFIG, STRING_CONFIG, BOOLEAN_CONFIG,
    LISTBOX_CONFIG, INTEGER_INPUT, DOUBLE_INPUT, STRING_INPUT, BOOLEAN_INPUT,
    FILE_UPLOAD, RADIO_BUTTONS, INTEGER_INPUT_LEGACY, DOUBLE_INPUT_LEGACY,
    STRING_INPUT_LEGACY, BOOLEAN_INPUT_LEGACY, FILE_UPLOAD_LEGACY,
    RADIO_BUTTONS_LEGACY
]

# Quickform nodes for which a text box has to be displayed
TEXTBOX_QF_NODES = [
    INTEGER_CONFIG, DOUBLE_CONFIG, STRING_CONFIG, INTEGER_INPUT, DOUBLE_INPUT,
    STRING_INPUT, INTEGER_INPUT_LEGACY, DOUBLE_INPUT_LEGACY,
    STRING_INPUT_LEGACY, LISTBOX_CONFIG
]

# The master list of Nodes supported
SUPPORTED_NODES = []
SUPPORTED_NODES.extend(ALLOWED_INPUT_NODES)
SUPPORTED_NODES.extend(ALLOWED_OUTPUT_NODES)
SUPPORTED_NODES.extend(ALLOWED_QF_NODES)
SUPPORTED_NODES.extend(DEPRECATED_INPUT_NODES)
SUPPORTED_NODES.extend(DEPRECATED_OUTPUT_NODES)

# Map of Input/output nodes to xml entry field, to specify the file name
NODE_TO_FIELD = {
    # Schrodinger Readers
    MOLECULE_READER: 'DataURL/0',
    ALIGNMENT_READER: 'DataURL',
    SEQUENCE_READER: 'DataURL',
    SMILES_READER: 'DataURL/0',
    PDB_READER: 'DataURL/0',
    FP_READER: 'filename',
    GRID_READER: 'DataURL/0',
    HYP_READER: 'DataURL/0',
    QSAR_READER: 'filename',
    CMS_READER: 'desmond_files/0',
    FEP_READER: 'DataURL/0',
    LD_INPUT: 'lr_sdf_file',
    # KNIME Readers
    SDF_READER: 'urls/0',
    MOL2_READER: 'filename',
    CSV_READER: 'settings/file_selection/path/path',
    EXCEL_READER: 'settings/file_selection/path/path',
    KNIME_TABLE_READER: 'settings/file_selection/path/path',
    KNIME_MODEL_READER: 'filechooser/path/path',
    KNIME_LIST_FILES: 'file_chooser/path/path',
    # Schrodinger Writers
    MOLECULE_WRITER: 'output_file_name',
    ALIGNMENT_WRITER: 'output_file_name',
    SEQUENCE_WRITER: 'output_file_name',
    FP_WRITER: 'fingerprint_writer_file_name',
    GRID_WRITER: 'output_file_name',
    HYP_WRITER: 'dest_directory',
    QSAR_WRITER: 'dest_directory',
    # KNIME Writers
    CSV_WRITER: 'settings/file_chooser_settings/path/path',
    SDF_WRITER: 'filename',
    MOL2_WRITER: 'output_file',
    EXCEL_WRITER: 'file_selection/path/path',
    KNIME_TABLE_WRITER: 'file_selection/path/path',
    # Deprecated Readers/Writers
    CSV_READER_1: 'url',
    EXCEL_READER_1: 'FILE_CHOOSER/path_or_url',
    KNIME_TABLE_READER_1: 'filename',
    KNIME_MODEL_READER_1: 'filename',
    KNIME_LIST_FILES_1: 'file_location',
    CSV_WRITER_1: 'filename',
    EXCEL_WRITER_1: 'FILE_CHOOSER/path_or_url',
    KNIME_TABLE_WRITER_1: 'filename'
}

# Handle Glide Grid Reader as we need to add two options:
# gridLocation and gridPrefix
GRID_LOCATION = 'gridLocation'
GRID_PREFIX = 'gridPrefix'
OLD_GRID_INPUT_FIELDS = [GRID_LOCATION, GRID_PREFIX]

# Global Regular expressions
INPUT_QUOTED = "%input_quoted%"
RE_SELECTED_ENTRIES = re.compile('%?selected entries%?', re.IGNORECASE)
RE_INCLUDED_ENTRIES = re.compile('%?included entries%?', re.IGNORECASE)
RE_INPUT_FILE = re.compile('%input%|%file%', re.IGNORECASE)
RE_FILE = re.compile('%file%', re.IGNORECASE)
RE_LIVEREPORT = re.compile('livereport', re.IGNORECASE)
RE_NEW_GROUP = re.compile('%?new group%?', re.IGNORECASE)
RE_MAESTRO_MESSAGE = re.compile('%?maestro message%?', re.IGNORECASE)
RE_MAESTRO_COMMANDS = re.compile('%?maestro commands%?', re.IGNORECASE)
RE_GUI_SETTING = re.compile(f'%input%|{INPUT_QUOTED}|%?guisetting%?',
                            re.IGNORECASE)
RE_GUI_SETTING_RADIO = re.compile('%?guisettingradio%?', re.IGNORECASE)
RE_GUI_SETTING_COMBO = re.compile('%?guisettingcombo%?', re.IGNORECASE)
# Deprecated ones
RE_STRUCT_FILE = re.compile('%file%|structure file', re.IGNORECASE)
RE_TEXT_FILE = re.compile('%file%|text file', re.IGNORECASE)

# Any node name - Basically to match all nodes irrespective of node name
RE_ANY = re.compile('.*')

# Option formats
# General option
OPTION_FORMAT = "-option=%s,%s,%s,%s"

# String option
STRING_OPTION_FORMAT = "-option=%s,%s,%s,String"

# Quick form nodes: key in settings.xml for variable name
# In different KNIME versions, different keys are used in settings.xml file
QF_VARIABLENAME_KEYS = ["variableName", "flowvariablename", "flowVariableName"]

# KNIME preference file path knime.pref
KNIME_PREF_FILE_NAME = "KNIME.pref"
MAESTRO_PRODUCT_DIR = "maestro" + \
    str(mmutil.mmfile_get_product_version("maestro")[1] // 1000)
KNIME_PREF_FILE = os.path.join(fileutils.get_directory_path(fileutils.APPDATA),
                               MAESTRO_PRODUCT_DIR, KNIME_PREF_FILE_NAME)

WORKFLOWSET_DOT_META = "workflowset.meta"


class KnimeWorkflowParser():

    def __init__(self, location, tempdir, logger):
        self.logger = logger
        self.logger.debug("Get some constants from the workflow...")

        self.workspaceDir = ""
        self.workflowName = ""

        if os.path.isdir(location):
            # location is the actual worlflow dir
            self.workspaceDir, self.workflowName = os.path.split(location)
        else:
            # location is the .py file path
            self.workspaceDir = os.path.splitext(location)[0] + "_dir"
            self.workflowName = os.path.split(os.path.splitext(location)[0])[1]

        if not (os.path.exists(self.workspaceDir)):
            raise ValueError("Specified workflow path:%s doesn't exist" % \
                    (self.workspaceDir))

        self.tmpDir = tempdir
        self.job_name = self.workflowName + "_" + ''.join(
            random.sample(string.ascii_letters + string.digits, 12))
        self.logger.debug("workflow : " +
                          os.path.join(self.workspaceDir, self.workflowName))
        self.logger.debug("SCHRO TEMPDIR :" + self.tmpDir)
        self.logger.debug("Jobname : " + self.job_name)
        self.script_name = __file__
        ### Use inspec.stack()[0][1] to avoid global name __File__ not define OR passed with location
        self.workflowDescription = ""
        self.input_nodes = []
        self.output_nodes = []
        self.GUI_parameters = []
        self.allowed_input_nodes = []
        self.allowed_output_nodes = []
        self.allowed_GUI_parameters = {}
        self.force = False
        self.startup_template = '''
__doc__ = """
Runs the KNIME workflow : %s

%s

Copyright Schrodinger, LLC. All rights reserved.
"""

import os, sys
from schrodinger.job import util
scripts_dir = os.path.join(os.environ['MMSHARE_EXEC'],
                           '..', '..', 'python', 'scripts')
sys.path.append(scripts_dir)

def install():
    from KNIME_workflow_install import workflow_install
    print("Running install in ", __file__)
    inst = workflow_install()
    inst.scenarios(__file__)

def run():
    from KNIME_workflow_wrapper import runWrapper
    print("Running run in :", __file__)
    runWrapper(__file__)

def update_tgz():
    print("Not for customer use")
    from KNIME_workflow_wrapper import update_tgz
    print("updating the tgz for :", __file__)
    update_tgz(__file__)
'''

    def lookUpTables(self):
        # Prepare the list of Allowed input nodes data
        for node in ALLOWED_INPUT_NODES + DEPRECATED_INPUT_NODES:
            self.allowed_input_nodes.append({
                'nodeName': RE_INPUT_FILE,
                'nodeType': node,
                'field': NODE_TO_FIELD[node]
            })

        # Now append the special entries
        # Selected entries
        self.allowed_input_nodes.append({
            'nodeName': RE_SELECTED_ENTRIES,
            'nodeType': MOLECULE_READER,
            'field': NODE_TO_FIELD[MOLECULE_READER]
        })
        #Included entries
        self.allowed_input_nodes.append({
            'nodeName': RE_INCLUDED_ENTRIES,
            'nodeType': MOLECULE_READER,
            'field': NODE_TO_FIELD[MOLECULE_READER]
        })
        # SDF Reader for LiveReport
        self.allowed_input_nodes.append({
            'nodeName': RE_LIVEREPORT,
            'nodeType': SDF_READER,
            'field': NODE_TO_FIELD[SDF_READER]
        })
        # CSV Reader for LiveReport
        self.allowed_input_nodes.append({
            'nodeName': RE_LIVEREPORT,
            'nodeType': CSV_READER,
            'field': NODE_TO_FIELD[CSV_READER]
        })

        # Now append the deprecated ones
        self.allowed_input_nodes.append({
            'nodeName': RE_STRUCT_FILE,
            'nodeType': MOLECULE_READER,
            'field': NODE_TO_FIELD[MOLECULE_READER]
        })
        self.allowed_input_nodes.append({
            'nodeName': RE_TEXT_FILE,
            'nodeType': CSV_READER_1,
            'field': NODE_TO_FIELD[CSV_READER_1]
        })
        # CSV Reader for LiveReport
        self.allowed_input_nodes.append({
            'nodeName': RE_LIVEREPORT,
            'nodeType': CSV_READER_1,
            'field': NODE_TO_FIELD[CSV_READER_1]
        })

        # 'LiveDesign Input' node
        self.allowed_input_nodes.append({
            'nodeName': RE_ANY,
            'nodeType': LD_INPUT,
            'field': NODE_TO_FIELD[LD_INPUT]
        })

        # Prepare the list of Allowed output nodes data
        for node in ALLOWED_OUTPUT_NODES + DEPRECATED_OUTPUT_NODES:
            self.allowed_output_nodes.append({
                'nodeName': RE_FILE,
                'nodeType': node,
                'field': NODE_TO_FIELD[node]
            })

        # Append the special entries
        # New group
        self.allowed_output_nodes.append({
            'nodeName': RE_NEW_GROUP,
            'nodeType': MOLECULE_WRITER,
            'field': NODE_TO_FIELD[MOLECULE_WRITER]
        })

        # Maestro Message
        for node in [CSV_WRITER, CSV_WRITER_1]:
            self.allowed_output_nodes.append({
                'nodeName': RE_MAESTRO_MESSAGE,
                'nodeType': node,
                'field': NODE_TO_FIELD[node]
            })

        # Maestro Commands
        for node in [CSV_WRITER, CSV_WRITER_1]:
            self.allowed_output_nodes.append({
                'nodeName': RE_MAESTRO_COMMANDS,
                'nodeType': node,
                'field': NODE_TO_FIELD[node]
            })

        # Append the deprecated ones
        self.allowed_output_nodes.append({
            'nodeName': RE_STRUCT_FILE,
            'nodeType': MOLECULE_WRITER,
            'field': NODE_TO_FIELD[MOLECULE_WRITER]
        })
        self.allowed_output_nodes.append({
            'nodeName': RE_TEXT_FILE,
            'nodeType': CSV_WRITER_1,
            'field': NODE_TO_FIELD[CSV_WRITER_1]
        })

        # Prepare the list of Allowed Quick form nodes data
        self.allowed_GUI_parameters = {\
            INTEGER_CONFIG: {'field': 'defaultValue/integer', 'type': 'int'},
            DOUBLE_CONFIG: {'field': 'defaultValue/double', 'type': 'double'},
            STRING_CONFIG: {'field': 'defaultValue/string', 'type': 'String'},
            BOOLEAN_CONFIG: {'field': 'defaultValue/boolean', 'type': 'boolean'},
            LISTBOX_CONFIG: {'field': 'defaultValue/string', 'type': 'String'},
            INTEGER_INPUT: {'field': 'defaultValue/integer', 'type': 'int'},
            DOUBLE_INPUT: {'field': 'defaultValue/double', 'type': 'double'},
            STRING_INPUT: {'field': 'defaultValue/string', 'type': 'String'},
            BOOLEAN_INPUT: {'field': 'defaultValue/boolean', 'type': 'boolean'},
            RADIO_BUTTONS: {'field': 'defaultValue/variable_value', 'type': 'String'},
            FILE_UPLOAD: {'field': 'defaultValue/path', 'type': 'String'},
            INTEGER_INPUT_LEGACY: {'field': 'value/value', 'type': 'int'},
            DOUBLE_INPUT_LEGACY: {'field': 'value/value', 'type': 'double'},
            STRING_INPUT_LEGACY: {'field': 'value/value', 'type': 'String'},
            BOOLEAN_INPUT_LEGACY: {'field': 'value/value', 'type': 'boolean'},
            RADIO_BUTTONS_LEGACY: {'field': 'value/value', 'type': 'String'},
            FILE_UPLOAD_LEGACY: {'field': 'value/location', 'type': 'String'}}

    def rebuild_location(self, node, metanodes):
        location = node['location'].split("/")
        new_location = ""
        for level in location:
            for number in metanodes.keys():
                if level == number:
                    if new_location == "":
                        new_location = metanodes[number][
                            "name"] + " (#" + level + ")"
                    else:
                        new_location = new_location + " - " + \
                          metanodes[number]["name"] + " (#" + level + ")"
        return new_location

    def getNodeId(self, settings_file):
        domparser = xml.dom.minidom.parse(settings_file)
        for entry in domparser.getElementsByTagName("entry"):
            if entry.getAttribute("key") == "factory":
                return entry.getAttribute("value")

    def parseWorkflow(self):
        self.logger.info("List input and output nodes...")
        workflow = os.path.join(self.workspaceDir, self.workflowName)
        self.logger.debug("Workflow: " + str(workflow))
        self.input_nodes = []
        self.output_nodes = []
        self.GUI_parameters = []
        metanodes = {}
        for root, subFolders, files in os.walk(workflow):
            node_name = ''
            # Node location in metanodes
            node_location = os.path.split(root)[0].replace(workflow + '/', '').\
              replace(workflow + '\\', '').replace(workflow, '')
            location = ''
            for level in node_location.split(')/'):
                if len(level.split('(#')) != 1:
                    if location == '':
                        location = level.split('(#')[1].replace(')', '')
                    else:
                        location = location + '/' + level.split(
                            '(#')[1].replace(')', '')
            node = os.path.split(root)[1]
            if ("(#" in node):  # Metanode or node
                nodeNumber = node.rsplit("(#")[1].replace(")", "")
                node = node.rsplit("(#")[0]
                node = node.strip()

                if "node.xml" not in files and "settings.xml" not in files:  # metanode
                    for xmlFile in files:
                        if xmlFile == "workflow.knime":
                            metanodes[nodeNumber] = {"name": "", "text": ""}
                            doc = xml.dom.minidom.parse(
                                os.path.join(root, xmlFile))
                            for entry in doc.getElementsByTagName("entry"):
                                if entry.getAttribute("key") == "name":
                                    if metanodes[nodeNumber][
                                            "name"] == "":  # first instance
                                        metanodes[nodeNumber]["name"] = \
                                          entry.getAttribute("value")
                                        self.logger.debug("metanode : " + nodeNumber + " - " + \
                                          metanodes[nodeNumber]["name"])
                                # Workflow not edited with KNIME 2.5
                                if entry.getAttribute("key") == "customName":
                                    metanodes[nodeNumber]["text"] = \
                                      entry.getAttribute("value")
                            config_nodes = doc.getElementsByTagName("config")
                            for config_node in config_nodes:
                                if config_node.getAttribute(
                                        "key") == 'nodeAnnotation':
                                    for entry in config_node.getElementsByTagName(
                                            "entry"):
                                        if entry.getAttribute("key") == "text":
                                            metanodes[nodeNumber]["text"] = \
                  entry.getAttribute("value").replace("%%00010", " ").replace("%%00013", " ")
                else:  # node
                    node_id = self.getNodeId(os.path.join(root, "settings.xml"))
                    if not node_id:
                        # KNIME-4788 - node_id could be None if settings.xml is
                        # from a wrapped metanode. So just ignore that
                        continue

                    self.logger.debug("NodeId:" + node_id)
                    # Parse only if node is in SUPPORTED_NODES list
                    if node_id not in SUPPORTED_NODES:
                        continue

                    self.logger.debug("Node location: " + str(location) +
                                      " - " + node_location)
                    self.logger.debug("Files : " + str(files) + " - " + nodeNumber + " - " + node \
                      + " - " + root)
                    # From KNIME 2.8, contents of node.xml are moved to settings.xml
                    # So, if there is no node.xml file in a node's directory,
                    # parse settings.xml itself to get the required parameters.
                    node_xml_file = "node.xml"
                    settings_xml_file = "settings.xml"
                    if "node.xml" not in files:
                        node_xml_file = "settings.xml"
                    file_name = ''
                    grid_location = ''
                    grid_prefix = ''
                    for xmlFile in files:
                        if xmlFile == settings_xml_file:  # Get node name
                            doc = xml.dom.minidom.parse(
                                os.path.join(root, xmlFile))
                            config_nodes = doc.getElementsByTagName("config")
                            for config_node in config_nodes:
                                if config_node.getAttribute(
                                        "key") == 'nodeAnnotation':
                                    for entry in config_node.getElementsByTagName(
                                            "entry"):
                                        if entry.getAttribute("key") == "text":
                                            node_name = entry.getAttribute("value").\
                                              replace("%%00010", " ").replace("%%00013", " ")

                            # Workflow not edited with KNIME 2.5
                            for entry in doc.getElementsByTagName("entry"):
                                if entry.getAttribute("key") == "customName":
                                    node_name = entry.getAttribute("value").\
                                      replace("%%00010", " ").replace("%%00013", " ")

                        if xmlFile == node_xml_file:
                            # Parse the node.xml mainly for the Quick form node/parameters
                            var_name = ""
                            var_value = ""
                            possible_choices = []
                            if os.path.exists(os.path.join(root, xmlFile)):
                                node_label = ''
                                node_description = ''
                                node_weight = 0
                                var_name = ''
                                var_value = ''
                                ### Or add warning if there is something missing?
                                doc = xml.dom.minidom.parse(
                                    os.path.join(root, xmlFile))

                                # Get the input/output field key to search. Also
                                # get the type: array field or not. If array
                                # field, check in '<config>' else in '<entry>'
                                field_key = ""
                                # For Quick form nodes, NODE_TO_FIELD will
                                # not have any entry for those nodes
                                is_array_field = False
                                if node_id in NODE_TO_FIELD:
                                    field_key = NODE_TO_FIELD[node_id]
                                    is_array_field = self.isArrayField(
                                        field_key)
                                    if is_array_field:
                                        # Remove '/0' at the end
                                        field_key = field_key[:-2]

                                for entry in doc.getElementsByTagName("entry"):
                                    attrib_key = entry.getAttribute("key")
                                    attrib_value = entry.getAttribute("value")

                                    # Get the file name
                                    if (not is_array_field) and \
                                        (attrib_key == os.path.basename(field_key)):
                                        file_name = attrib_value

                                    if attrib_key == "label":
                                        node_label = attrib_value
                                    if attrib_key == "description":
                                        node_description = attrib_value
                                    if attrib_key == "weight":
                                        node_weight = attrib_value
                                    if attrib_key == "value":
                                        var_value = attrib_value
                                    if attrib_key == "boolean":
                                        # KNIME-5619 For Boolean Input nodes,
                                        # the value will be in "boolean"
                                        var_value = attrib_value
                                    # For File Upload nodes, the value will be in "location"
                                    if attrib_key == "location":
                                        var_value = attrib_value
                                    # To support both Legacy and new QF nodes
                                    if attrib_key in QF_VARIABLENAME_KEYS:
                                        var_name = attrib_value

                                for config in doc.getElementsByTagName(
                                        "config"):
                                    attrib_key = config.getAttribute("key")
                                    if attrib_key in [
                                            "choices", "possible_choices"
                                    ]:
                                        for entry in config.getElementsByTagName(
                                                "entry"):
                                            if entry.getAttribute(
                                                    "key") != "array-size":
                                                possible_choices.append(
                                                    entry.getAttribute("value"))

                                    # Get the file_name
                                    if attrib_key == field_key:
                                        for entry in config.getElementsByTagName(
                                                "entry"):
                                            if entry.getAttribute("key") == "0":
                                                file_name = entry.getAttribute(
                                                    "value")

                                    # KNIME-4287
                                    # For Glide Grid Reader, look for older
                                    # values also for backward compatibility
                                    if attrib_key in OLD_GRID_INPUT_FIELDS:
                                        for entry in config.getElementsByTagName(
                                                "entry"):
                                            if entry.getAttribute("key") == "0":
                                                if attrib_key == GRID_LOCATION:
                                                    grid_location = entry.getAttribute(
                                                        "value")
                                                elif attrib_key == GRID_PREFIX:
                                                    grid_prefix = entry.getAttribute(
                                                        "value")
                                                break

                    self.logger.debug("Node : " + node + ", node number : " + str(nodeNumber) +\
                      ", Node name : " + node_name + ", Node location : " + node_location +\
                      ",  Location : " + location)

                    # Input nodes
                    # Handle Glide Grid Reader older values
                    # If new value not found but older values found, then port
                    # them to new one. Assuming always .zip extension
                    if node_id == GRID_READER and not file_name and \
                        (grid_location and grid_prefix):
                        file_name = os.path.join(grid_location,
                                                 grid_prefix) + ".zip"

                    for allowed in self.allowed_input_nodes:
                        pattern = allowed['nodeName']
                        if pattern.search(node_name) \
                            and (allowed['nodeType'] in node_id):
                            self.input_nodes.append({
                                'nodeNumber': nodeNumber,
                                'nodeType': node_id,
                                'nodeName': node_name,
                                'nodeLocation': node_location,
                                'fileName': file_name,
                                'location': location
                            })
                            break

                    # Output nodes
                    for allowed in self.allowed_output_nodes:
                        pattern = allowed['nodeName']
                        if pattern.search(node_name) \
                            and (allowed['nodeType'] in node_id):
                            self.output_nodes.append({
                                'nodeNumber': nodeNumber,
                                'nodeType': node_id,
                                'nodeName': node_name,
                                'nodeLocation': node_location,
                                'fileName': file_name,
                                'location': location
                            })
                            break

                    # GUI parameters
                    var_type = ""
                    input_quoted = False
                    for allowed in self.allowed_GUI_parameters:
                        if RE_GUI_SETTING_COMBO.search(node_name) and \
                            node_id in [RADIO_BUTTONS, RADIO_BUTTONS_LEGACY]:
                            var_type = 'COMBOBOX'
                            node_name = RE_GUI_SETTING_COMBO.sub('', node_name)
                        elif RE_GUI_SETTING_RADIO.search(node_name) and \
                            node_id in [RADIO_BUTTONS, RADIO_BUTTONS_LEGACY]:
                            var_type = 'RADIOBOX'
                            node_name = RE_GUI_SETTING_RADIO.sub('', node_name)
                        elif RE_GUI_SETTING.search(node_name) and \
                            (node_id in TEXTBOX_QF_NODES):
                            var_type = 'TEXTBOX'
                            input_quoted = INPUT_QUOTED in node_name
                            node_name = RE_GUI_SETTING.sub('', node_name)
                        elif RE_GUI_SETTING.search(node_name) and \
                            node_id in [FILE_UPLOAD, FILE_UPLOAD_LEGACY]:
                            var_type = 'FILEBROWSER'
                            node_name = RE_GUI_SETTING.sub('', node_name)
                        elif RE_GUI_SETTING.search(node_name) and \
                            node_id in [BOOLEAN_CONFIG, BOOLEAN_INPUT, BOOLEAN_INPUT_LEGACY]:
                            var_type = 'CHECKBOX'
                            node_name = RE_GUI_SETTING.sub('', node_name)

                        if var_type != "":
                            self.GUI_parameters.append({
                                'nodeName': node_name,
                                'nodeNumber': nodeNumber,
                                'nodeType': node_id,
                                'nodeLocation': node_location,
                                'location': location,
                                'nodeWeight': int(node_weight),
                                'nodeLabel': node_label,
                                'nodeDescription': node_description,
                                'input_quoted': input_quoted,
                                'parameters': [{
                                    'varName': var_name,
                                    'type': var_type,
                                    'possible values': possible_choices,
                                    'value': var_value
                                }]
                            })
                            self.logger.debug(str(self.GUI_parameters))
                            break

        # KNIME-3952
        # Sort input_nodes, output_nodes and GUI_parameters lists
        self.input_nodes.sort(key=lambda x: int(x['nodeNumber']))
        self.output_nodes.sort(key=lambda x: int(x['nodeNumber']))
        self.GUI_parameters.sort(key=lambda x: int(x['nodeNumber']))

        # List of the input, output nodes and GUI parameters
        self.logger.debug("Input node list : " + str(self.input_nodes))
        self.logger.debug("Output node list : " + str(self.output_nodes))
        for node in self.input_nodes:
            node['nodeLocation'] = self.rebuild_location(node, metanodes)
            self.logger.info("Input node : " + node['nodeName'] + " - " + node['nodeNumber'] + \
              " - " + node['nodeType'])
        for node in self.output_nodes:
            node['nodeLocation'] = self.rebuild_location(node, metanodes)
            self.logger.info("Output node : " + node['nodeName'] + " - " + node['nodeNumber'] + \
              " - " + node['nodeType'])
        for node in self.GUI_parameters:
            node['nodeLocation'] = self.rebuild_location(node, metanodes)
            self.logger.info("GUI parameters : " + node['nodeName'] + " - " + node['nodeNumber'] \
              + " - " + node['nodeLocation'] + " - " + node['location'] + " - " + node['nodeType'] \
              + " - " + str(node['nodeWeight']) + " - " + node['nodeLabel'] \
              + " - " + node['nodeDescription'])
            for param in node['parameters']:
                self.logger.info("                 " + param['varName'] + " - " + param['type'] + " - " + \
                  str(param['possible values']) + " - " + param['value'])

        self.workflowDescription, _ = KnimeWorkflowParser.getWorkflowDescription(
            workflow)
        if self.workflowDescription:
            self.logger.debug("Workflow decription : " +
                              self.workflowDescription)
        else:
            self.logger.debug("No workflow description")

    @staticmethod
    def getWorkflowDescription(workflow, tempdir=None):
        """
        Parse the workflowset.meta file passed and get the workflow description

        :param workflowset_file: Path to the workflowset.meta file under KNIME
            workflow
        :type workflowset_file: `str`

        :param tempdir: A temporary directory to extract the workflow
        :type tempdir: `str`

        :return: Return the workflow description and author
        :rtype: tuple(`str`, `str`)
        """
        workflow_description = ""
        author = ""
        is_extracted = False
        if os.path.isdir(workflow):
            workflowset_meta_file = os.path.join(workflow, WORKFLOWSET_DOT_META)
            if not os.path.isfile(workflowset_meta_file):
                return (workflow_description, author)
        elif os.path.isfile(workflow) and workflow.endswith('.knwf'):
            with ZipFile(workflow) as zf:
                workflow_name = zf.namelist()[0].split('/')[0]
                try:
                    workflowset_meta_file = zf.extract(
                        f'{workflow_name}/{WORKFLOWSET_DOT_META}', path=tempdir)
                except KeyError:
                    return (workflow_description, author)

                is_extracted = True
        else:
            return (workflow_description, author)

        doc = xml.dom.minidom.parse(workflowset_meta_file)
        workflowset_elements = doc.getElementsByTagName("element")
        # KNIME-5409
        # If there is any modification, workflowset.meta will contain
        # <met:element> tags instead of <element>
        if not workflowset_elements:
            workflowset_elements = doc.getElementsByTagName("met:element")

        for element in workflowset_elements:
            if element.getAttribute(
                    "name") == "Comments" and element.childNodes:
                workflow_description = element.childNodes[0].data

            if element.getAttribute("name") == "Author" and element.childNodes:
                author = element.childNodes[0].data

        if is_extracted:
            # Remove the extracted entry
            fileutils.force_rmtree(os.path.dirname(workflowset_meta_file))

        return (workflow_description, author)

    def isArrayField(self, field):
        return field.endswith("/0")

    def last_modified(self, workspace):
        self.logger.info(
            "Detecting the last modified workflow in the workspace: " +
            workspace)
        date_file_list = []
        wkf_list = glob.glob(os.path.join(workspace, '*', 'workflow.knime'))
        self.logger.debug("Workflow list : " + str(wkf_list))
        for workflow in wkf_list:
            stats = os.stat(workflow)
            lastmodDate = time.localtime(stats[8])
            fileDate = time.strftime("%m/%d/%y %H:%M:%S", lastmodDate)
            self.logger.debug(workflow + " modification date : " + fileDate)  ##
            date_file_tuple = lastmodDate, workflow
            date_file_list.append(date_file_tuple)
        self.logger.debug("Modification dates : " + str(date_file_list))
        date_file_list.sort()
        date_file_list.reverse()
        latest = date_file_list[0]
        fileName = os.path.split(os.path.split(latest[1])[0])[1]
        self.logger.debug("The latest modified workflow is : " + fileName)
        return fileName

    def name_synchro(self):
        self.logger.debug("Considering renaming the workspace...")
        new_name = self.last_modified(self.workspaceDir)
        edit = False  # Edit the workfow again if necessary
        in_scripts = maescripts.get_maestro_scripts_path(
            maescripts.USER_SCRIPTS_INSTALL_DIR_PATH)
        in_scripts = os.path.join(in_scripts, new_name + ".py")
        # If in scripts a provide another name to avoid interferences
        if " " in new_name:
            maestro.info("Please remove the space(s) in the workflow name: " +
                         new_name)
            edit = True
        elif new_name != self.workflowName:
            if os.path.exists(in_scripts) and not self.force:
                response = maestro.info("""There is already a workflow named:
%s
installed in Maestro under Scripts.
Please rename the current workflow in KNIME.
""" % (new_name.replace("_", " ")))
                edit = True
            else:
                response = maestro.question("""
The last saved or renamed workflow in the workspace is:
%s
Do you want to install this workflow in Scripts menu?
Install/Run - Redirects you to either install or run the workflow.
Reopen KNIME - Opens KNIME to edit/rename the workflow
              """ % (new_name.replace("_", " ")),
                                            button1="Install/Run",
                                            button2="Reopen KNIME")
                if response == maestro.BUTTON2:
                    edit = True
                else:
                    self.name_synchro_apply(new_name)
                    maestro.rebuild_scripts_menu()
                    edit = False
                    self.force = False
        elif self.force:
            self.logger.debug("Force mode")
            self.name_synchro_apply(new_name)
            maestro.rebuild_scripts_menu()
            edit = False
            self.force = False
        return edit

    def name_synchro_apply(self, new_name):
        new_workspace_name = self.workspaceDir.replace(self.workflowName,
                                                       new_name)
        self.logger.debug("Renaming the workfspace : " + self.workflowName \
          + " into " + new_workspace_name)
        if self.force:
            self.logger.debug(
                "Deleting existing workflow files corresponding to : " +
                new_workspace_name)
            if os.path.exists(new_workspace_name):
                shutil.rmtree(new_workspace_name)  ### Done later 838
            fileutils.force_remove(new_workspace_name.replace("_dir", ".py"))
            ##filetutils.force_remove(new_workspace_name.replace("_dir", ".pyc"))
            fileutils.force_remove(
                new_workspace_name.replace("_dir", "_install.py"))
            fileutils.force_remove(new_workspace_name.replace("_dir", ".tgz"))
            std_workflow_dir = maescripts.get_maestro_scripts_path(
                maescripts.STD_SCRIPTS_INSTALL_DIR_PATH)
            fileutils.force_remove(
                os.path.join(std_workflow_dir, new_name + ".tgz"))
            fileutils.force_remove(
                os.path.join(std_workflow_dir, new_name + "_install.py"))
        # Delete if already exists in the temporary directory
        if not self.force and ((self.tmpDir in new_workspace_name) \
          and os.path.exists(new_workspace_name)): # force or New
            shutil.rmtree(new_workspace_name)
            fileutils.force_remove(new_workspace_name.replace("_dir", ".py"))
        shutil.copytree(self.workspaceDir, new_workspace_name)
        self.logger.debug("Created : " + new_workspace_name)
        shutil.rmtree(self.workspaceDir)
        old_startup_name = self.workspaceDir.replace("_dir", ".py")
        if os.path.exists(old_startup_name):
            new_startup_name = new_workspace_name.replace("_dir", ".py")
            self.logger.debug("Editing the start up script :" +
                              new_startup_name)
            file2read = open(old_startup_name, 'r')
            destinationFile = open(new_startup_name, 'w')
            while 1:
                line = file2read.readline()
                if not line:
                    break
                line = line.replace(
                    self.workflowName.replace("_", " ") + "...",
                    new_name.replace("_", " ") + "...")
                line = line.replace("#Command:pythonrun " + \
                  self.workflowName + ".run",
                  "#Command:pythonrun " + new_name + ".run")
                ### build the lines again = more robust
                destinationFile.write(line)
            file2read.close()
            destinationFile.close()
            # temporary file clean-up
            fileutils.force_remove(old_startup_name)
            if os.path.exists(old_startup_name.replace(".py", ".pyc")):
                fileutils.force_remove(old_startup_name.replace(".py", ".pyc"))
            if not self.force:
                self.logger.debug("Editing the mnu file accordingly")
                mnu = maescripts.get_maestro_scripts_path(
                    maescripts.USER_SCRIPTS_MENU_PATH)
                mnu_tmp = os.path.join(self.tmpDir, self.job_name + ".mnu")
                file2read = open(mnu, 'r')
                # FIXME - Have to check if we are deleting other installed
                # scripts entry or not.
                # FIXME - This should be replaced by installscripts.py module
                # calls.
                destinationFile = open(mnu_tmp, 'w')
                while 1:
                    line = file2read.readline()
                    if not line:
                        break
                    line = line.replace(self.workflowName.replace("_", " ") \
                      + "...", new_name.replace("_", " ") + "...")
                    line = line.replace(
                        "pythonrun " + self.workflowName + ".run",
                        "pythonrun " + new_name + ".run")
                    ### build the lines again to be more robust
                    destinationFile.write(line)
                file2read.close()
                destinationFile.close()
                fileutils.force_rename(mnu_tmp, mnu)
                self.logger.debug("Altered the mnu file accordingly ")
        if not self.force:
            # New names
            self.workflowName = new_name
            self.workspaceDir = new_workspace_name
            self.logger.debug("The new workflow location is: " + \
            os.path.join(self.workspaceDir, self.workflowName))
        else:
            self.logger.debug("Updating mmshare...")
            std_workflow_dir = maescripts.get_maestro_scripts_path(
                maescripts.STD_SCRIPTS_INSTALL_DIR_PATH)
            fileutils.force_remove(
                os.path.join(new_workspace_name, ".metadata/knime/knime.log"))
            fileutils.force_remove(
                os.path.join(new_workspace_name,
                             ".metadata/knime/knime.log.old.gz"))
            tar = tarfile.open(
                os.path.join(std_workflow_dir, new_name + ".tgz"), "w:gz")
            tar.add(new_workspace_name, arcname=new_name + "_dir")
            tar.close()
            category = ":"  # Get workflow workflow category
            file2read = open(new_workspace_name.replace("_dir", ".py"), 'r')
            while 1:
                line = file2read.readline()
                if not line:
                    break
                if "#Name:" in line and line.count(':') == 2:
                    category = line.split(":")[1] + category
            file2read.close()
            wrapper_script = os.path.join(std_workflow_dir,
                                          new_name + "_install.py")
            # FIXME - This code must use installscripts.py to add a Name/Command pair
            # in the file.
            destinationFile = open(wrapper_script, 'w')
            txt = '#Name:' + category + new_name.replace("_", " ") \
              + ' - install/reset \n'
            txt = txt + '#Command:pythonrun ' + new_name + '_install.install \n'
            txt = txt + self.startup_template % (new_name.replace(
                "_", " "), self.workflowDescription)
            destinationFile.write(txt)
            destinationFile.close()
            self.logger.debug("tgz and _installed updated in mmshare")
            shutil.rmtree(new_workspace_name)
            fileutils.force_remove(new_workspace_name.replace("_dir", ".py"))

    def names_not_in_sync(self):
        self.logger.debug(
            "Checking the workflow and workspace names are in sync...")
        latest_name = self.last_modified(self.workspaceDir)
        workspace_name = os.path.split(self.workspaceDir.replace("_dir", ""))[1]
        self.logger.debug("The current workspace is: " + workspace_name)
        if latest_name != workspace_name:
            if maestro:
                maestro.info(
                    "Please give a more appropriate name to your workflow.")
            else:
                self.logger.warn(
                    "Please give a more appropriate name to your workflow.")
            return True
        else:
            return False

    def cmdBasis(self, save_after_exec):
        self.logger.debug("Creating the cmd basis...")
        run = os.path.join(os.environ['SCHRODINGER'], "run")
        cmd = [
            run, self.script_name, "-DISP", "ignore", self.job_name,
            os.path.join(self.workspaceDir, self.workflowName)
        ]

        if maestro:
            pt = maestro.project_table_get()
            cmd += ["-PROJ", pt.project_name]

        if not save_after_exec:
            cmd.append("-nosave")
        return cmd

    def checkNode(self, node, node_type, pattern):
        if (node_type in node['nodeType']) and pattern.search(node['nodeName']):
            return True
        else:
            return False

    def generateTempFile(self, file_suffix):
        random_str = ''.join(
            random.sample(string.ascii_letters + string.digits, 6))
        file_path = os.path.join(self.tmpDir,
                                 self.job_name + "_" + random_str + file_suffix)
        return file_path

    def _writeStructures(self, file_suffix, structures, message):
        '''
        Write the supplied structures to a temporary file and return the same.
        If number of structures is 0, returns an empty string
        '''
        file_name = ""
        num_structures = len(structures)
        self.logger.debug("Number of entries : " + str(num_structures))
        if num_structures == 0:
            maestro.command("showpanel table")
            maestro.warning(message)
        else:
            file_name = self.generateTempFile(file_suffix)
            mw = structure.MaestroWriter(file_name)
            for st in structures:
                mw.append(st)
            mw.close()

        return file_name

    def addInputData(self, cmd, in_maestro=True):
        self.logger.debug("Adding the input node options ...")
        for node in self.input_nodes:

            node_name = node['nodeName']
            node_type = node['nodeType']
            file_name = node['fileName']
            node_number = node['nodeNumber']

            self.logger.debug("%s - %s - %s" % (node_type, node_number, \
                        node_name))
            node_location = node['location']
            if node_location != '':
                node_location = node_location + "/"

            selected = self.checkNode(node, MOLECULE_READER,
                                      RE_SELECTED_ENTRIES)
            included = self.checkNode(node, MOLECULE_READER,
                                      RE_INCLUDED_ENTRIES)
            if maestro and (selected or included):
                # If Selected or included, then maestro has to be defined
                # Also, create a new temp file and write the
                # selected/included structures
                file_suffix = ""
                structures = []
                message = ""
                if (selected):
                    file_suffix = '_selected.mae'
                    pt = maestro.project_table_get()
                    rows = pt.selected_rows
                    for row in rows:
                        structures.append(row.getStructure())
                    message = "Please select at least one entry in " + \
                              "the Project Table"
                elif (included):
                    file_suffix = '_included.mae'
                    structures = maestro.get_included_entries()
                    message = "Please include at least one entry in " + \
                              "the workspace"

                file_name = self._writeStructures(file_suffix, structures,
                                                  message)
                if file_name == "":
                    return ""
                else:
                    node['fileName'] = file_name

            cmd_value = STRING_OPTION_FORMAT % \
                (node_location + node_number, \
                NODE_TO_FIELD[node_type], file_name)
            cmd.append(cmd_value)

            self.logger.debug("Adding: %s - %s - %s - %s to the input node " \
                    "list" % (node_type, node_number, node_name, file_name))

        self.logger.debug("Command : " + str(cmd))
        return cmd

    def addOutputData(self, cmd):
        self.logger.debug("Adding the output node options ...")
        i = 1
        for node in self.output_nodes:

            if maestro:
                new_group = self.checkNode(node, MOLECULE_WRITER, RE_NEW_GROUP)
                maestro_msg = self.checkNode(
                    node, CSV_WRITER, RE_MAESTRO_MESSAGE) or self.checkNode(
                        node, CSV_WRITER_1, RE_MAESTRO_MESSAGE)
                maestro_cmds = self.checkNode(
                    node, CSV_WRITER, RE_MAESTRO_COMMANDS) or self.checkNode(
                        node, CSV_WRITER_1, RE_MAESTRO_COMMANDS)

                if new_group:
                    node['fileName'] = self.generateTempFile(".mae")
                elif (maestro_msg or maestro_cmds):
                    node['fileName'] = self.generateTempFile(".csv")

            node_name = node['nodeName']
            node_type = node['nodeType']
            node_number = node['nodeNumber']
            file_name = node['fileName']

            self.logger.debug("%s - %s - %s" % (node_type, node_number, \
                        node_name))

            node_location = node['location']
            if node_location != '':
                node_location = node_location + "/"

            cmd_value = STRING_OPTION_FORMAT % \
                (node_location + node_number, \
                 NODE_TO_FIELD[node_type], file_name)
            cmd.append(cmd_value)
            i += 1  # make the file name unique

            self.logger.debug("Adding: %s - %s - %s - %s to the output node " \
                    "list" % (node_type, node_number, node_name, file_name))

        self.logger.debug("Command : " + str(cmd))
        return cmd

    def addGUIparameters(self, cmd):
        self.logger.debug("Adding the GUI parameters options ...")
        for node in self.GUI_parameters:
            node_location = node['location']
            if node_location != '':
                node_location = node_location + "/"
            for param in node[
                    'parameters']:  # several parameters per node in the future
                for key in self.allowed_GUI_parameters.keys():
                    allowed = self.allowed_GUI_parameters[key]
                    if (key in node['nodeType']):
                        cmd_value = OPTION_FORMAT % \
                            (node_location + node['nodeNumber'], \
                             allowed['field'], param['value'], allowed['type'])
                        cmd.append(cmd_value)
                        self.logger.debug("Adding : " + node['nodeType'] + " - " + \
                          node['nodeNumber'] + " - " + \
                          node['nodeName'] + " - to the GUI parameter list")
        self.logger.debug("Command : " + str(cmd))
        return cmd

    def run_monitor(self, cmd, show_monitor_panel):
        self.logger.debug("Run monitor calling: " +
                          str(subprocess.list2cmdline(cmd)) + "\n")

        jhandler = jobhandler.JobHandler(cmd)
        job = jhandler.launchJob()
        self.logger.info("SUBMITTED, waiting for the job to finish")

        if maestro and maestro.get_command_option(
                "prefer", "showmonitorpanel") and show_monitor_panel:
            maestro.command("showpanel monitor")

        jhandler.wait()
        # PANEL-19343
        # job object returned by launchJob() is not updated in place. So get the
        # updated job object once again here.
        job = jhandler.job
        if job.succeeded():
            self.logger.info("Job completed.")
        else:
            if maestro:
                maestro.warning(
                    "The KNIME workflow has failed; can't incorporate results")
            else:
                self.logger.error("The KNIME workflow has failed")

    def post_processing(self):
        self.logger.info("Post processing: Incorporating results in Maestro...")
        list1 = []
        list2 = []
        for node in self.output_nodes:  # First import the new structures
            if "Molecule Writer" in node['nodeType']:
                list1.append(node)
            else:
                list2.append(node)
        list1.extend(list2)
        for node in list1:
            for allowed in self.allowed_output_nodes:
                if self.checkNode(node, allowed['nodeType'], \
                        allowed['nodeName']):
                    self.logger.debug(node['nodeType'] + " - " + node['nodeNumber'] + " - " + \
                      node['nodeName'] + " - " + node['fileName'])
                    if not (os.path.isfile(node['fileName'])):
                        self.logger.debug("The output file name is: " + node['fileName'] \
                          + " is missing")
                        maestro.warning("KNIME output file is missing: %s" \
                          % node['fileName'])
                    else:
                        self.logger.debug("Output file name : " +
                                          node['fileName'])
                        if self.checkNode(node, MOLECULE_WRITER, RE_NEW_GROUP):
                            # New group in Project table
                            self.logger.info(
                                "Incorporating new structures in a group...")
                            # Rename group or file == _new_group for entry group naming
                            maestro.command(
                                'entryimport all=true extradata=true \
                              wsreplace=false fittoscreen=false creategroups=all \
                              "%s"' % (node['fileName']))

                        elif (self.checkNode(node, CSV_WRITER,
                                             RE_MAESTRO_MESSAGE) or
                              self.checkNode(node, CSV_WRITER,
                                             RE_MAESTRO_COMMANDS) or
                              self.checkNode(node, CSV_WRITER_1,
                                             RE_MAESTRO_MESSAGE) or
                              self.checkNode(node, CSV_WRITER_1,
                                             RE_MAESTRO_COMMANDS)):
                            # Maestro commands or message
                            file2read = open(node['fileName'], 'r')
                            message = ""
                            while 1:
                                line = file2read.readline()
                                if not line:
                                    break
                                if RE_MAESTRO_COMMANDS.search(node['nodeName']):
                                    maestro.command(line)
                                if RE_MAESTRO_MESSAGE.search(node['nodeName']):
                                    line = line.replace("$$$", "    ")
                                    message = message + line + "\n "
                            file2read.close()

                            if RE_MAESTRO_MESSAGE.search(node['nodeName']):
                                maestro.info(message)

                        elif self.checkNode(node, MOLECULE_WRITER,
                                            RE_STRUCT_FILE):
                            self.logger.info("Structure file")  ### FIX ME
                            fileutils.force_rename(
                                node['fileName'], node['fileName'].replace(
                                    self.tmpDir, os.getcwd()))

                        else:
                            maestro.info("File written to disk: " +
                                         node['fileName'])
                        break
        return

    def cleanup(self):
        if self.logger.getEffectiveLevel() != log.DEBUG:
            self.logger.debug("Clean up ...")
            #for node in self.input_nodes:
            #    fileutils.force_remove(node['fileName'])
            ##for node in output_nodes
            for tmpFile in glob.glob(
                    os.path.join(self.tmpDir, self.job_name + '.*')):
                fileutils.force_remove(tmpFile)
            ### But delete job_name based file sufficient?

            #### Run from Maestro ##########################################


def runWrapper(location):
    from KNIME_template_gui import KnimeTemplateGUI
    create_new = False
    logger = create_logger(location)

    # Create a temporary directory
    tempdir = tempfile.mkdtemp(prefix="KNIME_")
    logger.info("Temporary directory is: " + tempdir)

    if location == "NEW":  # Create a new workflow
        create_new = True
        location = copy_template_files("New_workflow", tempdir, logger)

    workflow_parser = KnimeWorkflowParser(location, tempdir,
                                          logger)  # Instance of the class

    if create_new or workflow_parser.names_not_in_sync():
        open_template(workflow_parser, logger)  # Open the new workflow

    # Allowed input, output and GUI parameter nodes
    workflow_parser.lookUpTables()
    # Parse workflow and create the node and parameter lists
    workflow_parser.parseWorkflow()
    workflowGUI = KnimeTemplateGUI(workflow_parser, create_new, logger)
    workflowGUI.show()
    workflowGUI.raise_()
    logger.info("Done")
    return workflowGUI, workflow_parser


def runNew():
    runWrapper("NEW")


def update_tgz(location):
    workflowGUI, workflow_parser = runWrapper(location)
    ##logger.info("")
    print("Update the tgz archive. Not for customer use.")
    workflowGUI.hide()
    workflow_parser.force = True
    workflowGUI.install_in_Maestro()
    ### skip Update mnu
    workflow_parser.workflowName = workflow_parser.workflowName + "_copy"
    workflow_parser.workspaceDir = workflow_parser.workspaceDir.replace(
        "_dir", "_copy_dir")
    workflowGUI.run_KNIME_GUI()


def copy_template_files(workflow_name, tempdir, logger):
    logger.info("Copy the template files...")
    ### workflowParser.tmpDir = tempfile.gettempdir() #### FIX ME
    location = os.path.join(tempdir, workflow_name)
    logger.debug("Workspace location: " + location + "_dir")  ##
    # Delete the old template files if there are some
    if os.path.exists(location + "_dir"):
        shutil.rmtree(location + "_dir")
    if os.path.exists(location + ".py"):
        fileutils.force_remove(location + ".py")
    if os.path.exists(location + ".pyc"):
        fileutils.force_remove(location + ".pyc")
    std_workflow_dir = maescripts.get_maestro_scripts_path(
        maescripts.STD_SCRIPTS_INSTALL_DIR_PATH)
    tar = tarfile.open(os.path.join(std_workflow_dir, workflow_name + ".tgz"))
    tar.extractall(tempdir)
    tar.close()
    return location


def create_logger(location, in_maestro=True, debug=False):
    script_basename = os.path.split(location)[1].replace(".pyc", "")
    logger = log.getLogger(script_basename)

    is_job_debug_enabled = 'False'
    if debug:
        is_job_debug_enabled = 'True'

    if maestro:
        mae_job_debug_pref = maestro.get_command_option("prefer",
                                                        "enablejobdebugoutput")
        is_job_debug_enabled = mae_job_debug_pref
        if is_job_debug_enabled == 'True':
            print("## Maestro preference 'enablejobdebugoutput' is",
                  mae_job_debug_pref)

    if is_job_debug_enabled == 'True':
        logger.setLevel(log.DEBUG)
    else:
        logger.setLevel(log.WARNING)

    if not logger.handlers:
        handler = log.StreamHandler(sys.stdout)
        handler.setFormatter(
            log.Formatter(script_basename + "::%(levelname)s::%(message)s"))
        logger.addHandler(handler)
        logger.debug("logger created")
    return logger


def open_template(workflowParser, logger):
    logger.info("Opening KNIME GUI to create a new workflow...")
    logger.info("Follow the directions listed in the template workflow.")
    ### ORIGINAL CODE
    pt = maestro.project_table_get()
    edit = True
    while edit:
        cmd = [
            os.path.join(os.environ['SCHRODINGER'],
                         "run"), workflowParser.script_name, "-PROJ",
            pt.project_name, "-DISP", "ignore", workflowParser.job_name,
            os.path.join(workflowParser.workspaceDir,
                         workflowParser.workflowName), "GUI"
        ]
        workflowParser.run_monitor(cmd, False)
        edit = workflowParser.name_synchro()


def parse_knime_pref():
    """
    Parse the APPDATA_DIR/maestroxxx/KNIME.pref file and returns a map of
    option to value. If an option does not have a value, then it will be set to
    None

    :return: map of option to value
    :rtype: map
    """
    user_option_map = {}
    if not os.path.isfile(KNIME_PREF_FILE):
        return user_option_map

    sys.stderr.write("Reading user options from '%s'" % KNIME_PREF_FILE)
    for line in open(KNIME_PREF_FILE, "r").readlines():
        if not line.strip() or line.startswith("#"):
            continue

        opt_arr = shlex.split(line)
        if len(opt_arr) == 1:
            user_option_map[opt_arr[0]] = None
        else:
            user_option_map[opt_arr[0]] = opt_arr[1]

    return user_option_map


def get_cmd_from_map(option_value_map):
    """
    Returns the command list from the option-value map. This simply appends all
    the option and value to the list and returns

    :param option_value_map: map of option-value
    :type  option_value_map: map

    :return: list of command options
    :rtype: list
    """
    cmd = []
    for option in list(option_value_map):
        cmd.append(option)
        value = option_value_map[option]
        if value:
            cmd.append(value)

    return cmd


def get_option_by_priority(knime_pref, cmd_option, default):
    """
    Returns the options value based on the priority. The order of priority is:
    value in KNIME.pref file -> value in command line -> default.
    The returned value can be None

    :param knime_pref: Value of the option in KNIME.pref file
    :type  knime_pref: str

    :param cmd_option: Value of the option in command line
    :type  cmd_option: str

    :param default: Default value for the option
    :type  default: str

    :return: The value of the option based on defined priority
    :rtype: str
    """
    option_value = default
    if knime_pref:
        option_value = knime_pref
    elif cmd_option:
        option_value = cmd_option

    return option_value


################# MAIN: Jobcontrol section ######################
if __name__ == '__main__':
    # This script is called 3 times:
    # - runWrapper(location)
    # - __main__ 1rst pass section
    # - __main__ 2nd pass section
    parser = argparse.ArgumentParser()

    parser.add_argument("-NOJOBID",
                        action="store_false",
                        default=True,
                        dest="use_jobcontrol",
                        help="Run the script without job control")
    parser.add_argument("-schrodingerTempDir",
                        action="store",
                        dest="schrod_temp_dir")
    parser.add_argument("-knimeTempDir", action="store", dest="knime_temp_dir")
    parser.add_argument("-knimeInstallDir",
                        action="store",
                        dest="knime_install_path")
    parser.add_argument("-printcmd",
                        action="store_true",
                        dest="printcmd",
                        default=False)

    (options, args) = parser.parse_known_args(sys.argv[1:])

    if options.use_jobcontrol:  # 1rst pass
        sys.stderr.write("Monitor the job using the Monitor Jobs panel... \n")
        job_name = args.pop(0)
        ### Use inspec.stack()[0][1] to avoid global name __File__ not define OR passed as argument
        if args[1] == 'GUI':
            prog = 'KNIME GUI'
        else:
            prog = 'KNIME'

        scriptlauncher = launcher.Launcher(script='KNIME_workflow_wrapper.py',
                                           prog=prog,
                                           jobname=job_name,
                                           wait=False,
                                           no_redirect=False)
        script_args = args[0:]

        # the backend will already be running under job control.
        script_args.append("-NOJOBID")

        # Resend the below options to the next run which are consumed by this
        # run
        if options.schrod_temp_dir:
            script_args.extend(['-schrodingerTempDir', options.schrod_temp_dir])
        if options.knime_temp_dir:
            script_args.extend(['-knimeTempDir', options.knime_temp_dir])
        if options.knime_install_path:
            script_args.extend(['-knimeInstallDir', options.knime_install_path])

        scriptlauncher.addScriptArgs(script_args)
        scriptlauncher.launch()

    else:  # 2nd pass
        if not options.printcmd:
            sys.stderr.write("Starting KNIME... \n")

        if (sys.platform == "win32"):
            knimeexec = os.path.join(os.environ['SCHRODINGER'], "knime.exe")
        else:
            knimeexec = os.path.join(os.environ['SCHRODINGER'], "knime")
        workflowToRun = args.pop(0)

        tmpDir = tempfile.gettempdir()
        knime_pref_map = parse_knime_pref()
        # Get the below options based on priority
        schrod_temp_dir = get_option_by_priority(
            knime_pref_map.pop('-schrodingerTempDir', None),
            options.schrod_temp_dir, tmpDir)
        knime_temp_dir = get_option_by_priority(
            knime_pref_map.pop('-knimeTempDir', None), options.knime_temp_dir,
            tmpDir)
        knime_install_dir = get_option_by_priority(
            knime_pref_map.pop('-knimeInstallDir', None),
            options.knime_install_path, "FromSchrodingerHost")

        cmd = [
            str(knimeexec), "-knimeTempDir", knime_temp_dir,
            "-schrodingerTempDir", schrod_temp_dir, "-knimeInstallDir",
            knime_install_dir, "-verbose"
        ]

        # Append the remaining options from knime_pref_map
        cmd.extend(get_cmd_from_map(knime_pref_map))

        if len(args) > 0 and args[0] == "GUI":
            sys.stderr.write("DEBUG	Workflow: " + workflowToRun + "\n")

            cmd.extend(["-data", os.path.split(workflowToRun)[0], "&"])
            sys.stderr.write(
                "STATUS	Opening the workflow in KNIME for edition... \n \n")
            sys.stderr.write("STATUS	" + subprocess.list2cmdline(cmd) + "\n \n")
            subprocess.call(cmd)
            sys.stderr.write("\nSTATUS	The KNIME GUI has been closed \n")
        else:
            if not options.printcmd:
                sys.stderr.write("STATUS	Running KNIME in batch... \n")
                sys.stderr.write("DEBUG	Workflow: " + workflowToRun + "\n")

            if os.path.isdir(workflowToRun):
                workflowToRun = "-workflowDir=" + workflowToRun
            elif workflowToRun.endswith(".zip") or \
                workflowToRun.endswith(".knwf"):
                workflowToRun = "-workflowFile=" + workflowToRun

            cmd.extend(["-batch", "-nosplash", workflowToRun])

            # Append the remaining arguments so that those will be simply passed
            # to Knime startup
            cmd.extend(args)

            for cmdArg in args:
                if cmdArg.startswith('-option='):
                    # Ex: "-option=17,value/value,view1,view2,view3,String"
                    # Ex: "-option=17,DataURL/0,/tmp/small_molecule.mae,String"
                    opt_arr = cmdArg.replace("-option=", "").split(",")
                    message = opt_arr[0] + " - " + \
                              ",".join(opt_arr[2:len(opt_arr) - 1])
                    if not options.printcmd:
                        sys.stderr.write("ARGUMENT	" + message + "\n")
                else:
                    if not options.printcmd:
                        sys.stderr.write("ARGUMENT	" + cmdArg + "\n")

            if not options.printcmd:
                sys.stderr.write("STATUS	Running KNIME: " +
                                 subprocess.list2cmdline(cmd) + "\n \n")
            else:
                sys.stderr.write("KNIME batch Command:\n" +
                                 subprocess.list2cmdline(cmd) + "\n")

            if options.printcmd:
                sys.exit(0)

            ret_code = subprocess.call(cmd)
            sys.stderr.write("\nSTATUS :	finished command line \n")
            sys.exit(ret_code)
