import argparse
import csv
import json
import os
import sys
from collections import OrderedDict

from schrodinger.structure import SDReader
from schrodinger.structure import SDWriter
from schrodinger.structure import StructureReader
from schrodinger.structure import StructureWriter
from schrodinger.utils import subprocess

PARAMS_JSON_FILE = 'params.json'
# keys in params.json file
DEPENDENT_COLUMN_DETAILS = 'dependent_column_details'
COLUMN_NAME = 'name'
VALUE_TYPE = 'value_type'
MODEL = 'model'
ID = 'id'
LIVEREPORT_ID = 'livereport_id'

# Internal column types (as in params.json)
TYPE_3D = '3D'
TYPE_STRING = 'string'
TYPE_REAL = 'float'
TYPE_INTEGER = 'integer'
TYPE_BOOLEAN = 'string'
TYPE_IMAGE = 'image'
TYPE_OTHER = 'attachment'

LR_COLTYPE_TO_MAE_PROPTYPE = {
    TYPE_STRING: 's',
    TYPE_REAL: 'r',
    TYPE_INTEGER: 'i',
    TYPE_BOOLEAN: 's',
    TYPE_IMAGE: 's',
    TYPE_OTHER: 's'
}

# Different 3D column types
TYPE_3D_LIGAND = 'LIGAND'
TYPE_3D_DOCKED_POSE = 'DPPOSE'
TYPE_3D_RL_COMPLEX = 'RL_COMPLEX'

# Column names in csv files
CORPORATE_ID = 'Corporate ID'
LIGAND = 'Ligand'
PROTEIN = 'Protein'
DATA = 'Data'

# Corporate id property in sdf file
CORPORATE_ID_PROPERTY = 's_sd_Corporate_ID'

MODEL_ID_FILE = 'model_id.txt'
LIVEREPORT_ID_FILE = 'livereport_id.txt'
EXTRA_STRUCT_ARGS_FILE = 'extra_struct_args.txt'
NUM_3D_COLUMNS_SUPPORTED = 1

# This file will be read/parsed by 'LiveDesign Input' node
INPUT_SPECS = 'input_specs.json'
ORIG_NAME = 'orig_name'

DESCRIPTION = '''
This script will get the individual properties from the list of csv fies and
add them as properties to the input sdf.
'''


def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION,
        prog="$SCHRODINGER/run multicolumn_to_sdf.py")

    parser.add_argument("-help",
                        action="store_true",
                        dest="help",
                        help="Show this help message and exit")

    parser.add_argument("-input", required=True, help='Input SDF file')

    parser.add_argument("-multicolumns",
                        nargs='+',
                        help='Multiple input column CSVs from the LiveReport')

    parser.add_argument(
        "-output",
        help='''Output SDF file. By default output will be written to
        <input>-out.sdf''')

    return parser.parse_args()


def write_model_and_lr_ids(params_json_file):
    with open(params_json_file, 'r') as fp:
        params = json.load(fp)
        with open(MODEL_ID_FILE, 'w') as model_writer:
            model_writer.write(params[MODEL][ID])

        with open(LIVEREPORT_ID_FILE, 'w') as lr_writer:
            lr_writer.write(params[LIVEREPORT_ID])


def get_column_details(params_json_file):
    columns = None
    with open(params_json_file, 'r') as fp:
        params = json.load(fp)
        columns = params[DEPENDENT_COLUMN_DETAILS]

    return columns


def convert_pse_to_mae(pse_file):
    pymol_exec = os.getenv('PYMOL_EXEC')
    if not pymol_exec:
        print('PYMOL_EXEC not set. Could not convert pse file to mae file.')
        return None

    basename, ext = os.path.splitext(pse_file)
    out_maefile = f'{basename}.mae'

    pml_file = f'{basename}.pml'
    with open(pml_file, 'w') as fw:
        fw.write(f'cmd.load("{pse_file}")')
        fw.write(f'\ncmd.save("{out_maefile}")')

    cmd = [pymol_exec, '-c', pml_file]
    subprocess.call(cmd)

    return out_maefile if os.path.isfile(out_maefile) else None


def guess_three_d_type(column_data):
    '''
    Guess the 3D type (like simple ligands, pose viewer, rl complex) for the
    given cell column data

    ligands - If only ligands available without protein
    docked poses - Same protein file for all ligand poses
    rl complex - each ligand has different protein

    Sample 3D column data:
    ligand:
    {'value_type': '3D', 'Data': [('ligand_1.mae', ''), ('ligand_2.mae', '')]}

    docked poses:
    {'value_type': '3D', 'Data': [('ligand_1.mae', 'protein.mae'), ('ligand_2.mae', 'protein.mae')]}

    rl complex
    {'value_type': '3D', 'Data': [('ligand_1.mae', 'protein_1.mae'), ('ligand_2.mae', 'protein_2.mae')]}
    '''
    proteins = set()
    for _, protein in column_data[DATA]:
        if protein:
            proteins.add(protein)

    three_d_type = None
    if len(proteins) > 1:
        # Multiple unique proteins in the set
        three_d_type = TYPE_3D_RL_COMPLEX
    elif len(proteins) == 1:
        # Only one unique protein in the set
        three_d_type = TYPE_3D_DOCKED_POSE
    else:
        three_d_type = TYPE_3D_LIGAND

    return three_d_type


def write_3d_file(column_data, output_file):
    '''
    Write a concatenated structure file for all the ligands and proteins in the
    given 3D column_data. First the proteins are written followed by ligands.

    Sample 3D column data:
    {'value_type': '3D', 'Data': [('Row5.mae', 'Row0.mae'), ('Row7.mae', 'Row0.mae')]}

    :param column_data: A dictionary representing a livereport 3D cell
    :type column_data: `dict`

    :return: Tuple of ligand file and protein file
    :rtype: tuple(str, str)
    '''
    struct_writer = StructureWriter(output_file)

    def append_3d_file(struct_file, struct_writer):
        if os.path.splitext(struct_file)[1] == '.pse':
            struct_file = convert_pse_to_mae(struct_file)

        if struct_file:
            for st in StructureReader(struct_file):
                struct_writer.append(st)

    three_d_type = guess_three_d_type(column_data)
    if three_d_type == TYPE_3D_LIGAND:
        # Write all the ligands in one file
        for ligand_file, _ in column_data[DATA]:
            append_3d_file(ligand_file, struct_writer)
    elif three_d_type == TYPE_3D_RL_COMPLEX:
        # Write protein, ligand, protein, ligand etc... (rl complexes)
        for ligand_file, protein_file in column_data[DATA]:
            append_3d_file(protein_file, struct_writer)
            append_3d_file(ligand_file, struct_writer)
    elif three_d_type == TYPE_3D_DOCKED_POSE:
        # Write the single protein followed by all the ligands (pv format)
        protein_written = False
        for ligand_file, protein_file in column_data[DATA]:
            if not protein_written:
                append_3d_file(protein_file, struct_writer)
                protein_written = True

            append_3d_file(ligand_file, struct_writer)

    struct_writer.close()


def file_not_empty(inp_file):
    '''
    Return True if file exisis and not empty else return False
    '''
    return os.path.isfile(inp_file) and os.path.getsize(inp_file) > 0


def populate_prop_dicts(csv_file, prop_name, prop_type, prop_dicts):
    '''
    Populates the `prop_dicts` by reading the column data from csv_file. The
    csv_file contains the data for a particular column for all rows.

    Sample csv_file contents:

    For non 3D columns:

    Corporate ID,Data
    V51993,-5.20601142973438
    V63720,-5.73352047269639

    For 3D columns:

    Corporate ID,Ligand,Protein
    V51993,Row5.mae,Row0.mae
    V51993,Row7.mae,Row0.mae
    V63720,Row4.mae,Row0.mae

    {

        CORP_ID: {prop_1: {'value_type': '3D', 'Data': [('ligand_1.mae',
                         'protien_1.mae'), ('ligand_2.mae', 'protein_1.mae')]},

                  prop_2: {'value_type': 'float', 'Data': '-5.2'}}

    }

    Sample property dict written by this method -
    {

        'V51993': {'column1': {'value_type': '3D', 'Data': [('Row5.mae',
                'Row0.mae'), ('Row7.mae', 'Row0.mae')]},
                'column2': {'value_type': 'float', 'Data': '-5.20601142973438'}},
        'V63720': {'column1': {'value_type': '3D', 'Data': [('Row4.mae',
                'Row0.mae')]},
                'column2': {'value_type': 'float', 'Data': '-5.73352047269639'}}

    }
    '''

    with open(csv_file, 'r') as fd:
        csv_reader = csv.DictReader(fd)
        for row in csv_reader:
            corp_id = row[CORPORATE_ID]

            if corp_id not in prop_dicts.keys():
                # prop_dicts doesn't contain corp_id yet, so add an entry
                prop_dicts[corp_id] = OrderedDict()

            if prop_name not in prop_dicts[corp_id].keys():
                # corp_id dict doesn't contain prop_name yet, so add an entry
                # DATA is a list of tuples of protein and ligand if type is 3D
                # else a simple value
                data = [] if prop_type == TYPE_3D else None
                prop_dicts[corp_id][prop_name] = {
                    VALUE_TYPE: prop_type,
                    DATA: data
                }

            if prop_type == TYPE_3D:
                ligand_file = row.get(LIGAND)
                protein_file = row.get(PROTEIN)
                prop_dicts[corp_id][prop_name][DATA].append(
                    (ligand_file, protein_file))
            else:
                prop_dicts[corp_id][prop_name][DATA] = row[DATA]


def parse_column_data(csv_files):
    '''
    Each csv file is written by livedesign on selecting a column as input in a
    parametrized model. Each of these files contain a single property/column
    value for all rows, like,

    For non 3D columns:

    Corporate ID,Data
    1234,10
    1235,11

    For 3D columns:

    Corporate ID,Ligand,Protein
    1234,lig_file_1.mae,prot_file_1.mae
    1235,lig_file_2.mae,prot_file_2.mae

    Parse the column data stored in the csv files and return a property dict

    For each column, a csv file is written with file name <id>.csv. This <id>
    can be mapped to the livereport column by looking up in params.json file
    'dependent_column_details' section.

    :param csv_files: List of property files written by livedesign on selecting a
        column as input
    :type csv_file: list(str)

    :return: Return column(property) values for all rows, like dict of corporate
        id to a dict of column(property) name to value
    :rtype: dict(str, dict(str, str))
    '''

    # Get the column details from params.josn
    columns_dict = get_column_details(PARAMS_JSON_FILE)

    prop_dicts = {}

    column_specs = []

    for count, csv_file in enumerate(csv_files):
        prop_id = os.path.splitext(csv_file)[0]

        # Get the column name and type from columns_dict
        # itself
        column = columns_dict.get(prop_id, None)
        prop_type = column[VALUE_TYPE] if column else None
        orig_prop_name = column[COLUMN_NAME]
        prop_name = f'column{count + 1}'

        populate_prop_dicts(csv_file, prop_name, prop_type, prop_dicts)

        # Add the property/column specs
        column_specs.append({
            COLUMN_NAME: prop_name,
            ORIG_NAME: orig_prop_name,
            VALUE_TYPE: prop_type
        })

    input_specs = {'columns': column_specs}
    # Write the column specs
    with open(INPUT_SPECS, 'w') as fd:
        json.dump(input_specs, fd, indent=1)

    return prop_dicts


def update_sdf(sdf_file, column_files, out_file=None):
    '''
    Update the input sdf file by adding properties for each csv column file
    passed

    :param sdf_file: Input sdf file
    :type sdf_file: `str`

    :param column_files: List of csv column files
    :type column_files: `list(`str`)`
    '''
    if not out_file:
        base, ext = os.path.splitext(sdf_file)
        out_file = f'{base}_modified{ext}'

    prop_dicts = parse_column_data(column_files)

    sd_writer = SDWriter(out_file)
    for structure in SDReader(sdf_file):
        corp_id = structure.property[CORPORATE_ID_PROPERTY].strip()
        prop_dict = prop_dicts.get(corp_id)
        if prop_dict:
            for prop_name, prop_data in prop_dict.items():
                if prop_data[VALUE_TYPE] == TYPE_3D:
                    # 3D type
                    struct_file = f'{corp_id}_{prop_name}.mae'
                    write_3d_file(prop_data, struct_file)
                    structure.property[f's_sd_{prop_name}'] = struct_file
                else:
                    # Non 3D type
                    value = prop_data[DATA]
                    if value:
                        # TODO: Write prop type char based on value type?
                        structure.property[f's_sd_{prop_name}'] = value

        sd_writer.append(structure)

    sd_writer.close()


if __name__ == '__main__':
    args = parse_args()

    sdf_file = args.input
    column_files = args.multicolumns

    if not column_files:
        sys.exit(0)

    update_sdf(sdf_file, column_files, args.output)
    write_model_and_lr_ids(PARAMS_JSON_FILE)
