import argparse
import copy
import csv
import json
from json import JSONDecodeError
import sys
import os
import tempfile
import traceback
import warnings

from schrodinger import structure
from schrodinger.application.livedesign import entry_types
from schrodinger.job.util import hunt
from schrodinger.utils import subprocess
from schrodinger.utils import fileutils
from schrodinger.utils.env import prepend_sys_path

with prepend_sys_path(os.path.join(hunt('knime'), '..', '..', 'python')):
    import knime_ld_utils as klu

# Ignore deprecation warnings while executing pymol commands
warnings.simplefilter('ignore', DeprecationWarning)

DESCRIPTION = """
Utility script that is run as part of KNIME protocols to post process the
results.csv file written by 'Upload as LiveDesign Model' node. Not
intended to be used outside.

If csvfile is for type docked poses:
'ligand_idx' column of second row (first row is header) contains the receptor
file. This script removes the second row which contains the receptor file and
adds the receptor filename in 'protein' column of all rows

If csvfile is for type rl complex:
The 'ligand_idx' column should contain filenames of rl complex structures.
This script scans through the 'ligand_idx' column of each row, reads the complex
structures, splits them into protein and ligands based on the entry_type and
write them into individual files. Then it replaces the 'ligand_idx' and
'protein' columns with the respective filenames.
"""

ROW_ID_COLUMN = 'RowId'
CORPORATE_ID_COLUMN = 'Corporate ID'
PROTEIN_COLUMN_NAME = 'protein'
# PyMOL throws warning if object name is 'protein'. So if column name is
# 'protein', the pymol object name will be 'receptor'
PROTEIN_OBJECT_NAME = 'receptor'
ENTRY_TYPES_WITH_TARGET = [
    klu.DOCKED_POSES, klu.RECEPTOR_LIGAND_COMPLEXES,
    klu.COVALENT_DOCKING_COMPLEXES
]

# Column types
STRING = 'String'
DOUBLE = 'Number (double)'
INTEGER = 'Number (integer)'
MAESTRO = 'Maestro'
SURFACE = 'Surface'
HYPOTHESIS = 'Hypothesis'

PYMOL_EXEC = os.getenv('PYMOL_EXEC')
PYMOL_EXEC_NOT_SET_MSG = 'PYMOL_EXEC is not set'

RESULTS_FILE = "results.csv"
SPECS_FILE = "specs.json"
USER_PYMOL_SCRIPT = "user_script.pml"
# The workflow name used in the protocol in KNIME_batch.py command for -destFile
# option
EXECUTED_WORKFLOW_NAME = 'executed.knwf'


def parse_args():
    # Parses the command line arguments
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION,
        prog="$SCHRODINGER/run postproces_ld_results.py")

    parser.add_argument(
        "-results_file",
        action="store",
        default=RESULTS_FILE,
        help="""Specify the results.csv file path written by 'Upload as LiveDesign
        Model' node. Default is results.csv from CWD""")

    parser.add_argument(
        "-specs_file",
        action="store",
        default=SPECS_FILE,
        help="""Specify the specs.json file path written by 'Upload as LiveDesign
        Model' node. Default is specs.json from CWD""")

    parser.add_argument(
        "-no_knwf",
        action="store_true",
        help=
        """Don't keep the executed workflow (if it is too large for instance). 
        Default is to keep it for debugging.""")

    parser.add_argument(
        "-user_pymol_script",
        action="store",
        default=USER_PYMOL_SCRIPT,
        help="""PyMOL script file containing user pymol commands (optional). If
        available (written by 'Upload as LiveDesign Model' node), these commands
        will be executed while creating pse file for surfaces. Default is
        user_script.pml from CWD""")

    return parser.parse_args()


class Specs:
    """
    Parse a specs.json file
    Format::

        {
            "entry_type":"rl_complex",
            "columns":[
                        {"column_name":"CT","column_type":"Maestro"},
                        {"column_name":"Corporate ID","column_type":"String"},
                        {"column_name":"docking score","column_type":"Number (double)"},
                        {"column_name":"Surface","column_type":"Surface"}
                      ]
        }
    """
    ENTRY_TYPE = "entry_type"
    COLUMNS = "columns"
    COLUMN_NAME = 'column_name'
    COLUMN_TYPE = 'column_type'

    def __init__(self, specs_file):
        """
        throws JSONDecodeError, KeyError
        """
        self.specs_file = specs_file
        self.entry_type = None
        self.ligand_column = None
        self.surface_columns = []
        self.hypo_columns = []
        self.has_row_id = False

        self.readSpecs()

    def readSpecs(self):
        """
        throws JSONDecodeError, KeyError
        """
        with open(self.specs_file, 'r') as sp:
            specs = json.load(sp)
            self.entry_type = specs[Specs.ENTRY_TYPE]
            columns = specs[Specs.COLUMNS]
            for column in columns:
                if column[Specs.COLUMN_TYPE] == MAESTRO:
                    self.ligand_column = column[Specs.COLUMN_NAME]
                elif column[Specs.COLUMN_TYPE] == SURFACE:
                    self.surface_columns.append(column[Specs.COLUMN_NAME])
                elif column[Specs.COLUMN_TYPE] == HYPOTHESIS:
                    self.hypo_columns.append(column[Specs.COLUMN_NAME])
                elif column[Specs.COLUMN_NAME] == ROW_ID_COLUMN:
                    self.has_row_id = True


class Postprocessor:

    def __init__(self, results_file, specs_file, user_pymol_script=None):
        self.results_file = results_file
        self.specs = Specs(specs_file)
        self.user_pymol_script = user_pymol_script

        self.pv_receptor_file = None

    def getRowkey(self, row):
        """
        Return the value of ROW_ID column if available else the value of
        CORPORATE_ID_COLUMN
        """
        return row[ROW_ID_COLUMN] if self.specs.has_row_id else row[
            CORPORATE_ID_COLUMN]

    @staticmethod
    def get_object_name(col_name):
        if col_name == PROTEIN_COLUMN_NAME:
            return PROTEIN_OBJECT_NAME
        else:
            return col_name.replace(' ', '_')

    @staticmethod
    def valid_file(filepath):
        return filepath and os.path.isfile(
            filepath) and os.path.getsize(filepath) > 0

    def processHeader(self, header):
        processed_header = []

        for name in header:
            if name == self.specs.ligand_column:
                processed_header.append(name)
                if self.specs.entry_type in ENTRY_TYPES_WITH_TARGET:
                    # If ligand column and has target, add a protein column next to ligand column
                    processed_header.append(PROTEIN_COLUMN_NAME)
            elif name in self.specs.surface_columns:
                # Skip all surface columns
                continue
            else:
                processed_header.append(name)

        return processed_header

    def processRowForDockedPose(self, row):
        """
        Process a csv row from a results.csv file of type docked pose

        :param row: A csv row read by csv.DictReader
        :type row: `dict`

        :return: Returns the modified row or None
        :rtype: `dict` or None

        """
        if not self.pv_receptor_file:
            # If pv receptor file is not yet fetched, then it must be the first
            # row in docked poses output. First row must contain the receptor
            # file. Just store the receptor file stored in the ligand column
            # This row should not be added to processed rows, so retuen False
            self.pv_receptor_file = row[self.specs.ligand_column]
            return False

        # If type is docked poses, then write the receptor_file in
        # protein column
        row[PROTEIN_COLUMN_NAME] = self.pv_receptor_file
        return True

    def processRowForRLComplex(self, row):
        """
        Process a csv row from a results.csv file of type rl complex and returns
        one or more rows as list

        :param row: A row read by csv.DictReader
        :type row: dict

        :return: Return a list of rows
        :rtype: list(dict)
        """
        # Ligand column must contain a rl complex file
        complex_file = row[self.specs.ligand_column]
        complex_file_basename, complex_file_ext = os.path.splitext(complex_file)

        if not complex_file:
            # Empty cell. Just add an empty protein column and return the same
            # row as a list
            row[PROTEIN_COLUMN_NAME] = ''
            return [row]

        structures = list(structure.StructureReader(complex_file))

        rl_map = klu.get_receptor_ligand_map(structures, entry_types.Complexes)

        rows = []

        # Create a default row by cleaning up the ligand and target columns,
        # basically the 3D data. The ligands and targets are appended to the
        # row from the rl_map
        default_row = copy.copy(row)
        default_row[self.specs.ligand_column] = ''
        default_row[PROTEIN_COLUMN_NAME] = ''

        # Add ligands and receptors from rl_map
        for idx, rl_group in enumerate(rl_map.rl_groups):
            modified_row = copy.copy(default_row)
            if rl_group.ligand:
                ligand_file = f'{complex_file_basename}_ligand_{idx}{complex_file_ext}'
                rl_group.ligand.write(ligand_file)
                modified_row[self.specs.ligand_column] = ligand_file

            if rl_group.receptor:
                receptor_file = f'{complex_file_basename}_receptor_{idx}{complex_file_ext}'
                rl_group.receptor.write(receptor_file)
                modified_row[PROTEIN_COLUMN_NAME] = receptor_file

            if (modified_row[self.specs.ligand_column] or
                    modified_row[PROTEIN_COLUMN_NAME]):
                # Append the row only if we have a ligand or receptor
                rows.append(modified_row)

        # If rows is empty, then we did not get any ligands or receptors from
        # the complex. So we should add the default row so that the other non 3D
        # column values are added to the result
        if not rows:
            rows.append(copy.copy(default_row))

        return rows

    def processRowForSurfaces(self, row, column=None):
        """
        Create a pse by applying all the surfaces (in all surface columns) on
        either ligand or protein column (as specified in column argument). This
        pse file then replaces the value in specified column.
        If column is not specified, then ligand column will be considered.

        :param row: Row dict as read by csv.DictReader
        :type row: `dict`

        :param column: Column name of either ligand or protein column. Default
                       is ligand column
        :type column: `str`
        """
        if not self.specs.surface_columns:
            # No surface columns. So just return
            return

        # Remove all the surface columns from row and store the surface col
        # names and files as a list of tuples for further processing
        surfaces = [(col, row.pop(col)) for col in self.specs.surface_columns]

        if not PYMOL_EXEC:
            # Surface columns are present but PYMOL_EXEC is not set. We cannot
            # create the pse file. So just return
            print(f'{PYMOL_EXEC_NOT_SET_MSG}. Could not apply Surfaces.')
            return

        if not column:
            # Default is ligand column
            column = self.specs.ligand_column

        struct_file = row[column]
        if not struct_file:
            # If file is missing, then return
            return

        # Create the pse file with all the surface column and replace the
        # column with this pse file
        row_key = self.getRowkey(row)

        if self.specs.has_row_id:
            pse_file = f'{row_key}_Surface.pse'
        else:
            _, pse_file = tempfile.mkstemp(prefix=f'{row_key}_surface_',
                                           suffix='.pse',
                                           dir=os.getcwd())
            pse_file = os.path.basename(pse_file)

        _, pml_file = tempfile.mkstemp(prefix=f'{row_key}_',
                                       suffix='.pml',
                                       dir=os.getcwd())
        pml_file = os.path.basename(pml_file)

        with open(pml_file, 'w') as pml_fw:
            # Load the ligand/protein file, with column name as object name
            obj_name = Postprocessor.get_object_name(column)
            pml_fw.write(f"cmd.load('{struct_file}', '{obj_name}')\n")
            for column_name, surface_file in surfaces:
                # KNIME-5452
                # Load with the column name as object name for this surface object.
                # This way, additional pymol commands can be added by user to
                # control the surface using the column name as object name.
                obj_name = Postprocessor.get_object_name(column_name)
                pml_fw.write(f"cmd.load('{surface_file}', '{obj_name}')\n")

            # Append user pymol script/commands
            self.appendUserPyMOLScript(pml_fw)
            pml_fw.write(f"cmd.save('{pse_file}')\n")

        status = Postprocessor.run_pymol_script(pml_file)
        if status and Postprocessor.valid_file(pse_file):
            row[column] = pse_file
        else:
            print(f'Error in creating pse file for surfaces for row {row_key}')

    def appendUserPyMOLScript(self, pml_fw):
        if not Postprocessor.valid_file(self.user_pymol_script):
            return

        with open(self.user_pymol_script, 'r') as fh:
            for line in fh.readlines():
                pml_fw.write(line)

    def createPseForHypothesisColumns(self, row):
        hypotheses = [(col, row[col]) for col in self.specs.hypo_columns]

        row_key = self.getRowkey(row)
        for column_name, hypo_file in hypotheses:
            if not Postprocessor.valid_file(hypo_file):
                continue

            hypo_name = os.path.splitext(hypo_file)[0]
            pse_file = hypo_name + '.pse'
            if Postprocessor.valid_file(pse_file):
                # File already exist for this hypothesis
                row[column_name] = pse_file
                continue

            # Create a pse file for hypothesis
            _, pml_file = tempfile.mkstemp(prefix=f'{row_key}_',
                                           suffix='.pml',
                                           dir=os.getcwd())
            pml_file = os.path.basename(pml_file)
            with open(pml_file, 'w') as pml_fw:
                obj_name = Postprocessor.get_object_name(column_name)
                pml_fw.write(f"cmd.load('{hypo_file}', '{obj_name}')\n")
                pml_fw.write("cgos = cmd.get_names_of_type('object:cgo')\n")
                pml_fw.write("xvol = cmd.get_names(selection='*_xvol')\n")
                pml_fw.write("cgos += xvol\n")
                pml_fw.write(f"cmd.save('{pse_file}', ' '.join(cgos))\n")

            status = Postprocessor.run_pymol_script(pml_file)
            if status and Postprocessor.valid_file(pse_file):
                row[column_name] = pse_file
            else:
                row[column_name] = ''
                print(
                    f'Error in creating pse file for Hypothesis. row:{row_key} column:{column_name}'
                )

    def createLigandPseForHypothesis(self, row):
        ligand_file = row[self.specs.ligand_column]

        # Create the pse file for ligand column and replace the ligand
        # column with this pse file
        row_key = self.getRowkey(row)

        if self.specs.has_row_id:
            pse_file = f'{row_key}_Hypothesis.pse'
        else:
            _, pse_file = tempfile.mkstemp(prefix=f'{row_key}_ligand_hypo_',
                                           suffix='.pse',
                                           dir=os.getcwd())
            pse_file = os.path.basename(pse_file)

        _, pml_file = tempfile.mkstemp(prefix=f'{row_key}_',
                                       suffix='.pml',
                                       dir=os.getcwd())

        pml_file = os.path.basename(pml_file)
        with open(pml_file, 'w') as pml_fw:
            pml_fw.write("set transparency_mode, 3\n")
            pml_fw.write(f"cmd.load('{ligand_file}')\n")
            ligand_file_basename = os.path.splitext(ligand_file)[0]
            pml_fw.write(f"cmd.set('valence', 1, '{ligand_file_basename}')\n")
            pml_fw.write(
                f"cmd.set('sphere_transparency', 0.0, '{ligand_file_basename}')\n"
            )
            pml_fw.write(f"cmd.save('{pse_file}')\n")

        status = Postprocessor.run_pymol_script(pml_file)
        if status and Postprocessor.valid_file(pse_file):
            row[self.specs.ligand_column] = pse_file
        else:
            print(
                f'Error in creating ligand pse file for hypothesis for row {row_key}'
            )

    def processRowForHypothesis(self, row):
        """
        Replace Hypothesis columns with pse files
        """
        if not self.specs.hypo_columns:
            # No Hypothesis columns. So just return
            return

        if not PYMOL_EXEC:
            # Hypothesis columns are present but PYMOL_EXEC is not set. We cannot
            # create the pse file. So just return
            print(f'{PYMOL_EXEC_NOT_SET_MSG}. Could not apply Hypothesis.')
            return

        self.createPseForHypothesisColumns(row)

        ligand_file = row[self.specs.ligand_column]
        if not (ligand_file and row[self.specs.hypo_columns[0]]):
            # If ligand file or first hypothesis column is missing, then return
            return

        self.createLigandPseForHypothesis(row)

    def convertMaeToPse(self, mae_file):
        """
        Convert mae_file to pse format and return the pse file

        :param mae_file: A mae file
        :type mae_file: `str`

        :return: pse file if conversion success else None
        :rtyep: `str` or None
        """
        basename = os.path.splitext(mae_file)[0]
        pse_file = f'{basename}.pse'
        if Postprocessor.valid_file(pse_file):
            # pse file already exists for this mae.
            return pse_file

        _, pml_file = tempfile.mkstemp(prefix=f'{basename}_',
                                       suffix='.pml',
                                       dir=os.getcwd())
        pml_file = os.path.basename(pml_file)

        with open(pml_file, 'w') as pml_fw:
            pml_fw.write(f"cmd.load('{mae_file}')\n")
            pml_fw.write(f"cmd.save('{pse_file}')\n")

        status = Postprocessor.run_pymol_script(pml_file)
        if not (status and Postprocessor.valid_file(pse_file)):
            print(f'Error in converting {mae_file} to {pse_file}')
            return None

        return pse_file

    def matchLigandTargetFileTypes(self, row):
        """
        LDIDEAS-5023
        Ligand and target file types must be same. Ligand and target file types
        could be one of .mae and .pse. In that case, convert the .mae to .pse
        """
        if not PROTEIN_COLUMN_NAME in row.keys():
            # If protein column is not found then return
            return

        ligand_file = row[self.specs.ligand_column]
        target_file = row[PROTEIN_COLUMN_NAME]

        ligand_ext = os.path.splitext(ligand_file)[1]
        target_ext = os.path.splitext(target_file)[1]

        if ligand_ext == target_ext:
            # Both types are .mae
            return

        # If control reaches here, then one of those is .mae and another one is
        # .pse. So convert the .mae to .pse
        if ligand_ext == '.mae':
            ligand_pse = self.convertMaeToPse(ligand_file)
            if ligand_pse:
                row[self.specs.ligand_column] = ligand_pse
        else:
            # target is .mae
            target_pse = self.convertMaeToPse(target_file)
            if target_pse:
                row[PROTEIN_COLUMN_NAME] = target_pse

    def writeCSVRow(self, csv_writer, row):
        """
        Write a csv row to the the given DictWriter
        :param csv_writer: An csv.DictWriter instance
        :type csv_writer: `csv.DictWriter`

        :param row: A row read by csv.DictReader
        :type row: `dict`
        """
        self.matchLigandTargetFileTypes(row)
        csv_writer.writerow(row)

    @staticmethod
    def run_pymol_script(pml_file):
        command = [PYMOL_EXEC, '-c', pml_file]
        completed_proc = subprocess.run(command, cwd=os.getcwd())
        if completed_proc.returncode != 0:
            print(
                f'Error in executing pymol script, Return code:{completed_proc.returncode}'
            )
            return False

        return True

    def run(self):
        """
        Process the results.csv based on the specs
        """
        # Moving results.csv to orig_results.csv and processing each row in
        # orig_results.csv, and writing the processed row in results.csv
        orig_results_file = f'orig_{os.path.basename(self.results_file)}'
        os.rename(self.results_file, orig_results_file)

        with open(orig_results_file, newline='') as inp_csv:
            csv_reader = csv.DictReader(inp_csv)
            processed_header = self.processHeader(csv_reader.fieldnames)

            with open(self.results_file, 'w', newline='') as out_csv:
                csv_writer = csv.DictWriter(out_csv,
                                            fieldnames=processed_header)
                csv_writer.writeheader()

                for row in csv_reader:
                    if self.specs.entry_type == klu.DOCKED_POSES:
                        if self.processRowForDockedPose(row):
                            self.processRowForSurfaces(row)
                            # Assuming Docked poses cannot contain Hypothesis column
                            # Do we need to call processRowForHypothesis(row)?
                            self.writeCSVRow(csv_writer, row)
                    elif self.specs.entry_type == klu.RECEPTOR_LIGAND_COMPLEXES:
                        rl_rows = self.processRowForRLComplex(row)
                        for rl_row in rl_rows:
                            self.processRowForSurfaces(rl_row,
                                                       PROTEIN_COLUMN_NAME)
                            # Assuming rl complex cannot have Hypothesis column
                            # Do we need to call this processRowForHypothesis(rl_row)?
                            self.writeCSVRow(csv_writer, rl_row)
                    elif self.specs.entry_type == klu.LIGANDS:
                        self.processRowForSurfaces(row)
                        self.processRowForHypothesis(row)
                        self.writeCSVRow(csv_writer, row)
                    else:
                        print(
                            f'Entry type {self.specs.entry_type} not supported.'
                        )
                        raise ValueError


def add_executed_workflow(results_file):
    """
    Append an additional column in results.csv to store the path to the executed
    workflow
    """
    # Rename results.csv to orig_results.csv and then write the modified
    # content to results.csv
    orig_results_file = f'orig_{os.path.basename(results_file)}'
    os.rename(results_file, orig_results_file)

    new_column_name = 'Executed workflow'
    with open(orig_results_file, newline='') as inp_csv:
        csv_reader = csv.DictReader(inp_csv)
        processed_header = csv_reader.fieldnames + [new_column_name]

        with open(results_file, 'w', newline='') as out_csv:
            csv_writer = csv.DictWriter(out_csv, fieldnames=processed_header)
            csv_writer.writeheader()

            for row in csv_reader:
                row[new_column_name] = EXECUTED_WORKFLOW_NAME
                csv_writer.writerow(row)


if __name__ == '__main__':
    args = parse_args()

    results_file = args.results_file
    specs_file = args.specs_file

    # KNIME-5869 Add additional column to store executed workflow
    if not args.no_knwf:
        add_executed_workflow(results_file)

    if not os.path.isfile(specs_file):
        print(f'{specs_file} file does not exist. Skipping post process.')
        sys.exit(0)

    if not Postprocessor.valid_file(results_file):
        print(f'{os.path.abspath(results_file)} file does not exist or empty.')
        sys.exit(1)

    try:
        post_processor = Postprocessor(results_file,
                                       specs_file,
                                       user_pymol_script=args.user_pymol_script)
    except (JSONDecodeError, KeyError):
        print('Error in creating Postprocessor instance.')
        traceback.print_exc()
        fileutils.force_remove(results_file)
        sys.exit(1)

    try:
        post_processor.run()
    except Exception as ex:
        traceback.print_exc()
        fileutils.force_remove(results_file)
