import argparse
import os

from schrodinger.utils.env import prepend_sys_path
from schrodinger.utils import subprocess
from schrodinger.utils import fileutils

with prepend_sys_path(os.environ['MMSHARE_EXEC']):
    import structconvert

DESCRIPTION = '''
Preprocess KNIME protocol

- Rename the files specified in extra_file_x options to the filename specified
in extra_arg_x option.
Example:
$SCHRODINGER/run preprocess_protocol.py -extra_file_1='oldfile.txt'
-extra_arg_1='-istruct newfile.txt'

For the above command, oldfile.txt will be renamed to newfile.txt

If -extra_arg_1='-dummy', the renaming will be ignored.

Note: Since the value for -extra_arg_1 starts with a '-', specifying
-extra_arg_1 '-dummy' fails. So always specify -extra_arg_1='-dummy'
'''

DEFAULT_ARG_VALUE = '-dummy'
SCHRODINGER = os.environ['SCHRODINGER']
SCHRODINGER_UTILITIES = os.path.join(SCHRODINGER, 'utilities')


def parse_args():
    """
    Parse the command line arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION,
        prog="$SCHRODINGER/run preprocess_protocol.py")

    parser.add_argument("-help",
                        action="store_true",
                        dest="help",
                        help="Show this help message and exit")

    parser.add_argument("-extra_arg_1", help='Extra argument 1')
    parser.add_argument("-extra_arg_2", help='Extra argument 2')
    parser.add_argument("-extra_arg_3", help='Extra argument 3')
    parser.add_argument("-extra_file_1", help='Extra file 1')
    parser.add_argument("-extra_file_2", help='Extra file 2')
    parser.add_argument("-extra_file_3", help='Extra file 3')

    return parser.parse_args()


def get_filename_from_arg(arg):
    '''
    Parse the source file name from the extra argument
    '''
    if not arg:
        return None

    arg = arg.strip()
    if not arg or arg == DEFAULT_ARG_VALUE or not arg.startswith('-'):
        return None

    try:
        opt, filename = arg.split()
    except ValueError:
        return None

    if '.' not in filename:
        return None

    return filename


def do_convert(source_file, dest_file):
    '''
    Check if conversion is required
    '''
    if fileutils.get_file_extension(
            source_file.lower()) == fileutils.get_file_extension(
                dest_file.lower()):
        return False

    return bool(
        structconvert.find_format(source_file) and
        structconvert.find_format(dest_file))


def rename_or_convert(source_file, dest_file):
    if source_file == dest_file:
        return

    if do_convert(source_file, dest_file):
        # Convert
        print(f'Converting "{source_file}" to "{dest_file}"')
        command = [
            os.path.join(SCHRODINGER_UTILITIES, 'structconvert'), source_file,
            dest_file
        ]
        ret_code = subprocess.call(command)
    else:
        # Rename
        print(f'Renaming "{source_file}" to "{dest_file}"')
        os.rename(source_file, dest_file)


def rename_input_files(extra_arg1=None,
                       extra_arg2=None,
                       extra_arg3=None,
                       extra_file1=None,
                       extra_file2=None,
                       extra_file3=None):

    for extra_arg, source_file in [(extra_arg1, extra_file1),
                                   (extra_arg2, extra_file2),
                                   (extra_arg3, extra_file3)]: # yapf: disable
        dest_file = get_filename_from_arg(extra_arg)
        if not (source_file and dest_file):
            continue

        rename_or_convert(source_file, dest_file)


if __name__ == '__main__':
    args = parse_args()

    rename_input_files(extra_arg1=args.extra_arg_1,
                       extra_arg2=args.extra_arg_2,
                       extra_arg3=args.extra_arg_3,
                       extra_file1=args.extra_file_1,
                       extra_file2=args.extra_file_2,
                       extra_file3=args.extra_file_3)
